"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

import numpy as np

from ANNarchy.intern.SpecificPopulation import SpecificPopulation
from ANNarchy.intern import Messages
from ANNarchy.core.Neuron import Neuron


class PoissonPopulation(SpecificPopulation):
    """
    Population of spiking neurons following a Poisson distribution.

    Each neuron of the population will randomly emit spikes, with a mean firing rate defined by the *rates* argument.

    The mean firing rate in Hz can be a fixed value for all neurons:

    ```python
    pop = net.create(ann.PoissonPopulation(geometry=100, rates=100.0))
    ```

    but it can be modified later as a normal parameter:

    ```python
    pop.rates = np.linspace(10, 150, 100)
    ```

    It is also possible to define a temporal equation for the rates, by passing a string to the argument:

    ```python
    pop = net.create(
        ann.PoissonPopulation(
            geometry=100, 
            rates="100.0 * (1.0 + sin(2*pi*t/1000.0) )/2.0"
        )
    )
    ```

    The syntax of this equation follows the same structure as neural variables.

    It is also possible to add parameters to the population which can be used in the equation of `rates`:

    ```python
    pop = net.create(
        ann.PoissonPopulation(
            geometry=100,
            parameters = dict(
                amp = 100.0,
                frequency = 1.0,
            ),
            rates="amp * (1.0 + sin(2*pi*frequency*t/1000.0) )/2.0"
        )
    )
    ```

    **Note:** The preceding definition is fully equivalent to the definition of this neuron:

    ```python
    poisson = ann.Neuron(
        parameters = dict(
            amp = 100.0,
            frequency = 1.0,
        ),
        equations = [
            'rates = amp * (1.0 + sin(2*pi*frequency*t/1000.0) )/2.0',
            'p = Uniform(0.0, 1.0) * 1000.0 / dt',
        ],
        spike = "p < rates"
    )
    ```

    The refractory period can also be set, so that a neuron can not emit two spikes too close from each other.

    If the ``rates`` argument is not set, the population can be used as an interface from a rate-coded population. The ``target`` argument specifies which incoming projections will be summed to determine the instantaneous firing rate of each neuron.

    ```python
    net = ann.Network()

    rates = 10.*np.ones((2, 100))
    rates[0, :50] = 100.
    rates[1, 50:] = 100.
    inp = net.create(ann.TimedArray(rates = rates, schedule=50.))

    pop = net.create(ann.PoissonPopulation(100, target="exc"))

    proj = net.connect(inp, pop, 'exc')
    proj.one_to_one(1.0)
    ```

    :param geometry: population geometry as tuple.
    :param name: unique name of the population (optional).
    :param rates: mean firing rate of each neuron. It can be a single value (e.g. 10.0) or an equation (as string).
    :param target: the mean firing rate will be the weighted sum of inputs having this target name (e.g. "exc").
    :param parameters: additional parameters which can be used in the `rates` equation.
    :param refractory: refractory period in ms.    
    """

    def __init__(self, 
                 geometry:int|tuple[int], 
                 name:str=None, 
                 rates:float|str=None, 
                 target:str=None, 
                 parameters:dict={}, 
                 refractory:float=None, 
                 copied:bool=False,
                 net_id:int = 0):

        if rates is None and target is None:
            Messages._error('A PoissonPopulation must define either rates or target.')

        self.target = target
        self.parameters_init = parameters
        self.refractory_init = refractory
        self.rates_init = rates

        if target is not None: # hybrid population
            # Create the neuron
            poisson_neuron = Neuron(
                parameters = parameters,
                equations = """
                    rates = sum(%(target)s)
                    p = Uniform(0.0, 1.0) * 1000.0 / dt
                    _sum_%(target)s = 0.0
                """ % {'target': target},
                spike = """
                    p < rates
                """,
                refractory=refractory,
                name="Hybrid",
                description="Hybrid spiking neuron emitting spikes according to a Poisson distribution at a frequency determined by the weighted sum of inputs."
            )


        elif isinstance(rates, str):
            # Create the neuron
            poisson_neuron = Neuron(
                parameters = parameters,
                equations = """
                    rates = %(rates)s
                    p = Uniform(0.0, 1.0) * 1000.0 / dt
                    _sum_exc = 0.0
                """ % {'rates': rates},
                spike = """
                    p < rates
                """,
                refractory=refractory,
                name="Poisson",
                description="Spiking neuron with spikes emitted according to a Poisson distribution."
            )

        elif isinstance(rates, np.ndarray):
            poisson_neuron = Neuron(
                parameters = """
                rates = 10.0
                """,
                equations = """
                p = Uniform(0.0, 1.0) * 1000.0 / dt
                """,
                spike = """
                p < rates
                """,
                refractory=refractory,
                name="Poisson",
                description="Spiking neuron with spikes emitted according to a Poisson distribution."
            )
        else:
            poisson_neuron = Neuron(
                parameters = """
                rates = %(rates)s
                """ % {'rates': rates},
                equations = """
                p = Uniform(0.0, 1.0) * 1000.0 / dt
                """,
                spike = """
                p < rates
                """,
                refractory=refractory,
                name="Poisson",
                description="Spiking neuron with spikes emitted according to a Poisson distribution."
            )
        SpecificPopulation.__init__(self, geometry=geometry, neuron=poisson_neuron, name=name, copied=copied, net_id=net_id)

        if isinstance(rates, np.ndarray):
            self.rates = rates

    def _copy(self, net_id=None):
        "Returns a copy of the population when creating networks."
        return PoissonPopulation(
            self.geometry, name=self.name, 
            rates=self.rates_init, target=self.target, 
            parameters=self.parameters_init, 
            refractory=self.refractory_init, 
            copied=True,
            net_id = self.net_id if not net_id else net_id,
        )

    def _generate_st(self):
        """
        Generate single thread code.

        We don't need any separate code snippets. All is done during the
        normal code generation path.
        """
        pass

    def _generate_omp(self):
        """
        Generate openMP code.

        We don't need any separate code snippets. All is done during the
        normal code generation path.
        """
        pass

    def _generate_cuda(self):
        """
        Generate CUDA code.

        We don't need any separate code snippets. All is done during the
        normal code generation path.
        """
        pass


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.generator.NanoBind.BaseTemplate import basetemplate
from ANNarchy.generator.NanoBind.Population import *
from ANNarchy.generator.NanoBind.Projection import *
from ANNarchy.generator.NanoBind.Profiler import profiler_template

from ANNarchy.intern.NetworkManager import NetworkManager
from ANNarchy.intern.ConfigManagement import ConfigManager, _check_paradigm
from ANNarchy.intern.GlobalObjects import GlobalObjectManager
from ANNarchy.parser.Extraction import extract_functions
from ANNarchy.intern.Profiler import Profiler

class NanoBindGenerator:
    """
    Create a Python-wrapper for the ANNarchy simulation core using the nanobind package (https://nanobind.readthedocs.io/en/latest/index.html)
    """
    def __init__(self, annarchy_dir, net_id):
        """
        Store a list of population und projection objects for later processing.
        """
        self._annarchy_dir = annarchy_dir
        self.net_id = net_id
        self.network = NetworkManager().get_network(self.net_id)

    def generate(self):
        """
        Generate the code stored in `ANNarchyCore[net_id].cpp`.
        """
        pop_struct_code = ""
        proj_struct_code = ""
        pop_mon_code = ""
        proj_mon_code = ""

        # Add device specific functions
        if _check_paradigm("openmp", self.net_id):
            device_specific = "\tm.def(\"set_number_threads\", &setNumberThreads);"
        elif _check_paradigm("cuda", self.net_id):
            device_specific = "\tm.def(\"set_device\", &setDevice);"
        else:
            raise NotImplementedError

        # Functions
        functions_code = self._generate_functions()

        # Constants
        constant_code = self._generate_constant(self.network.get_constants())

        # Populations
        for pop in self.network.get_populations():
            if 'wrapper' in pop._specific_template.keys():
                pop_struct_code += pop._specific_template['wrapper']
            else:
                pop_struct_code += self._generate_pop_wrapper(pop)

            if 'monitor_wrapper' in pop._specific_template.keys():
                pop_mon_code += pop._specific_template['monitor_wrapper']
            else:
                pop_mon_code += self._generate_pop_mon_wrapper(pop)

        # Projections
        for proj in self.network.get_projections():
            if 'wrapper' in proj._specific_template.keys():
                proj_struct_code += proj._specific_template['wrapper']
            else:
                proj_struct_code += self._generate_proj_wrapper(proj)

            if 'monitor_wrapper' in proj._specific_template.keys():
                proj_mon_code += proj._specific_template['monitor_wrapper']
            else:
                proj_mon_code += self._generate_proj_mon_wrapper(proj)

        # Profiling of simulation kernel is optional
        profiling_code = profiler_template if Profiler().enabled else ""

        return basetemplate % {
            'net_id': self.net_id,
            'device_specific': device_specific,
            'functions_wrapper': functions_code,
            'constant_wrapper': constant_code,
            'pop_struct_wrapper': pop_struct_code,
            'proj_struct_wrapper': proj_struct_code,
            'pop_mon_wrapper': pop_mon_code,
            'proj_mon_wrapper': proj_mon_code,
            'profiling_wrapper': profiling_code
        }

    def _generate_pop_wrapper(self, pop: "Population") -> str:
        """
        Generate wrapper for the C++ *PopStruct* structure.
        """
        additional_func = ""
        attributes = ""

        # synaptic delay
        attributes += """\t\t.def("update_max_delay", &PopStruct{id}::update_max_delay)\n""".format(id=pop.id)

        # Model attributes
        for attr in pop.neuron_type.description['attributes']:
            # internal variables should not exposed to Python
            if attr.startswith("rand_"):
                continue

            # Attributes are accessed directly from Python
            attributes += """\t\t.def_rw("{name}", &PopStruct{id}::{name})\n""".format(id=pop.id, name=attr)

            # On GPUs, we need to trigger a host-to-device transfer on next call of simulate()
            if _check_paradigm("cuda", self.net_id):
                # for local attributes
                if attr in pop.neuron_type.description['local']:
                    attributes += """\t\t.def_rw("{name}_host_to_device", &PopStruct{id}::{name}_host_to_device)\n""".format(id=pop.id, name=attr)
                # for global attributes which are updated over time
                if attr in pop.neuron_type.description['global'] and attr in [tmp['name'] for tmp in pop.neuron_type.description['variables']]:
                    attributes += """\t\t.def_rw("{name}_host_to_device", &PopStruct{id}::{name}_host_to_device)\n""".format(id=pop.id, name=attr)

        # Arrays for the post-synaptic potential
        if pop.neuron_type.type == 'rate':
            for target in sorted(list(set(pop.neuron_type.description['targets'] + pop.targets))):
                attributes += """\t\t.def_rw("{name}", &PopStruct{id}::{name})\n""".format(id=pop.id, name="_sum_"+target)

        # Functions
        for func in pop.neuron_type.description['functions']:
            # Wrapper
            additional_func += f"""
        .def("{func['name']}", &PopStruct{pop.id}::{func['name']})"""    
        

        # Type-specific functions
        if pop.neuron_type.type == "spike":
            additional_func += """\t\t.def("compute_firing_rate", &PopStruct{id}::compute_firing_rate)\n""".format(id=pop.id)

            if (pop.neuron_type.refractory or pop.refractory):
                additional_func += """\t\t.def_rw("refractory", &PopStruct{id}::refractory)\n""".format(id=pop.id)
                if _check_paradigm("cuda", self.net_id):
                    additional_func += """\t\t.def_rw("refractory_dirty", &PopStruct{id}::refractory_dirty)\n""".format(id=pop.id)

        wrapper_code = pop_struct_wrapper % {
            'id': pop.id,
            'attributes': attributes,
            'additional': additional_func
        }
        wrapper_code += '\n'
        return wrapper_code

    def _generate_proj_wrapper(self, proj: "Projection") -> str:

        methods = ""
        attributes = ""
        additional_func = ""

        # Connectivity as LIL
        connectivity = f"""
        .def("init_from_lil", &ProjStruct{proj.id}::init_from_lil)
        .def("post_rank", &ProjStruct{proj.id}::get_post_rank)
        .def("dendrite_size", &ProjStruct{proj.id}::dendrite_size)
        .def("nb_dendrites", &ProjStruct{proj.id}::nb_dendrites)
        .def("pre_ranks", &ProjStruct{proj.id}::get_pre_ranks)
        .def("pre_rank", &ProjStruct{proj.id}::get_dendrite_pre_rank)
        .def("nb_synapses", &ProjStruct{proj.id}::nb_synapses)"""

        # Special case for spiking projections
        if proj.synapse_type.type == "spike":
            connectivity += f"""
        .def("nb_efferent_synapses", &ProjStruct{proj.id}::nb_efferent_synapses)
"""

        # Contrary to *Population* attributes, we don't access the local attributes directly but use a common interface.
        datatypes = {
            'local': [],
            'semiglobal': [],
            'global': []
        }

        # Collect necessary data types
        for var in proj.synapse_type.description['parameters'] + proj.synapse_type.description['variables']:
            locality = var['locality']
            if var['name'] == "w" and proj._has_single_weight():
                locality = 'global'

            if var['ctype'] not in datatypes[locality]:
                datatypes[locality].append(var['ctype'])

        # Check if we need delay code
        has_delay = (proj.max_delay > 1)
        if proj.uniform_delay > 1 :
            key_delay = "uniform"
        else:
            if proj.synapse_type.type == "rate":
                key_delay = "nonuniform_rate_coded"
            else:
                key_delay = "nonuniform_spiking"
        if has_delay:
            attributes += proj_delays[key_delay] % {'id': proj.id}
        
        # Generate accessor for synapse attributes
        for ctype in datatypes["local"]:
            attributes += proj_local_attr % {'id': proj.id, 'ctype': ctype}

        for ctype in datatypes["semiglobal"]:
            attributes += proj_semiglobal_attr % {'id': proj.id, 'ctype': ctype}

        for ctype in datatypes["global"]:
            attributes += proj_global_attr % {'id': proj.id, 'ctype': ctype}

        # Functions
        for func in proj.synapse_type.description['functions']:
            # Wrapper
            additional_func += f"""
        .def("{func['name']}", &ProjStruct{proj.id}::{func['name']})"""  

        # Structural plasticity
        structural_plasticity = ConfigManager().get('structural_plasticity', self.net_id)
        if 'creating' in proj.synapse_type.description.keys():
            structural_plasticity = True
            attributes += f"""
        // Creating
        .def_rw("_creating", &ProjStruct{proj.id}::_creating)
        .def_rw("_creating_period", &ProjStruct{proj.id}::_creating_period)
        .def_rw("_creating_offset", &ProjStruct{proj.id}::_creating_offset)"""
            
        if 'pruning' in proj.synapse_type.description.keys():
            structural_plasticity = True
            attributes += f"""
        // Pruning
        .def_rw("_pruning", &ProjStruct{proj.id}::_pruning)
        .def_rw("_pruning_period", &ProjStruct{proj.id}::_pruning_period)
        .def_rw("_pruning_offset", &ProjStruct{proj.id}::_pruning_offset)"""

        if structural_plasticity:
            methods += f"""
        // Structural plasticity
        .def("synapse_exists", &ProjStruct{proj.id}::synapse_exists)
        .def("add_single_synapse", &ProjStruct{proj.id}::add_single_synapse)
        .def("add_multiple_synapses", &ProjStruct{proj.id}::add_multiple_synapses)
        .def("remove_single_synapse", &ProjStruct{proj.id}::remove_single_synapse)
        """
            
        # Generate the wrapper
        wrapper_code = proj_struct_wrapper % {
            'id': proj.id,
            'connectivity': connectivity,
            'methods': methods,
            'attributes': attributes,
            'additional': additional_func,
            'float_prec': ConfigManager().get("precision", self.net_id)
        }
        wrapper_code += '\n'
        return wrapper_code

    def _generate_pop_mon_wrapper(self, pop: "Population") -> str:
        """
        Generate wrapper for the C++ *PopRecorder* structure.
        """
        record_flag = ""
        record_container = ""
        clear_container = ""

        for attr in pop.neuron_type.description['variables']:
        
            record_flag +="""\t\t.def_rw("record_{name}", &PopRecorder{id}::record_{name})\n""".format(id=pop.id, name=attr['name'])
        
            record_container += """\t\t.def_rw("{name}", &PopRecorder{id}::{name})\n""".format(id=pop.id, name=attr['name'])
        
            clear_container += """\t\t.def("clear_{name}", &PopRecorder{id}::clear_{name})\n""".format(id=pop.id, name=attr['name'])

        # Arrays for the post-synaptic potential
        if pop.neuron_type.type == 'rate':
            for target in sorted(list(set(pop.neuron_type.description['targets'] + pop.targets))):
                record_flag += """\t\t.def_rw("record_{name}", &PopRecorder{id}::record_{name})\n""".format(id=pop.id, name="_sum_"+target)

                record_container += """\t\t.def_rw("{name}", &PopRecorder{id}::{name})\n""".format(id=pop.id, name="_sum_"+target)

                clear_container += """\t\t.def("clear_{name}", &PopRecorder{id}::clear_{name})\n""".format(id=pop.id, name="_sum_"+target)

        # Spike events        
        if pop.neuron_type.type == "spike":
        
            record_flag +="""\t\t.def_rw("record_{name}", &PopRecorder{id}::record_{name})\n""".format(id=pop.id, name='spike')
        
            record_container += """\t\t.def_rw("{name}", &PopRecorder{id}::{name})\n""".format(id=pop.id, name='spike')
        
            clear_container += """\t\t.def("clear_{name}", &PopRecorder{id}::clear_{name})\n""".format(id=pop.id, name='spike')

        wrapper_code = pop_mon_wrapper % {
            'id': pop.id,
            'record_flag': record_flag,
            'record_container': record_container,
            'clear_container': clear_container
        }
        wrapper_code += '\n'

        return wrapper_code


    def _generate_proj_mon_wrapper(self, proj: "Projection") -> str:
        """
        Generate wrapper for the C++ *PopRecorder* structure.
        """
        record_flag = ""
        record_container = ""
        clear_container = ""

        for attr in proj.synapse_type.description['variables']:
            
            record_flag +="""\t\t.def_rw("record_{name}", &ProjRecorder{id}::record_{name})\n""".format(id=proj.id, name=attr['name'])
            
            record_container += """\t\t.def_rw("{name}", &ProjRecorder{id}::{name})\n""".format(id=proj.id, name=attr['name'])
            
            clear_container += """\t\t.def("clear_{name}", &ProjRecorder{id}::clear_{name})\n""".format(id=proj.id, name=attr['name'])

        wrapper_code = proj_mon_wrapper % {
            'id': proj.id,
            'record_flag': record_flag,
            'record_container': record_container,
            'clear_container': clear_container
        }
        wrapper_code += '\n'

        return wrapper_code


    def _generate_constant(self, constants) -> str:
        """
        Generate wrapper for the C++ constants.
        """

        if len(constants) == 0:
            return ""

        wrapper_code = """
    // Constants"""
        
        for constant in constants:
            wrapper_code += f"""
    m.def("set_{constant.name}", &set_{constant.name});"""

        return  wrapper_code

    def _generate_functions(self) -> str:
        """
        Generate wrapper for the global functions.
        """

        if GlobalObjectManager().number_functions() == 0:
            return ""

        wrapper_code = """
    // Custom global functions"""
        
        for  _, function in GlobalObjectManager().get_functions():
            # Extract the name
            desc = extract_functions(description=function, local_global=True, net_id=self.net_id)[0]
            # Wrapper
            wrapper_code += f"""
    m.def("{desc['name']}", &{desc['name']});"""

        return  wrapper_code
    


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.intern.SpecificProjection import SpecificProjection
from ANNarchy.intern.ConfigManagement import ConfigManager, _check_paradigm
from ANNarchy.intern import Messages

from ANNarchy.core.PopulationView import PopulationView

class CurrentInjection(SpecificProjection):
    """
    Inject current from a rate-coded population into a spiking population.

    The pre-synaptic population must be be rate-coded, the post-synaptic one must be spiking, both must have the same size and no plasticity is allowed.

    For each post-synaptic neuron, the current `g_target` (e.g. `g_exc`or `g_inh`) will be set at each time step to the firing rate `r` of the pre-synaptic neuron with the same rank.

    The projection must be connected with `connect_current()`, which takes no parameter and does not accept delays. It is equivalent to `one_to_one(weights=1.0)`.

    Example:

    ```python
    inp = net.create(100, ann.Neuron(equations="r = 5*sin(t/1000)"))
    pop = net.create(100, ann.Izhikevich)

    proj = net.connect(ann.CurrentInjection(inp, pop, 'exc'))
    proj.connect_current()
    ```

    :param pre: pre-synaptic population.
    :param post: post-synaptic population.
    :param target: type of the connection.
    :param name: optional name.

    """
    def __init__(self, pre:"Population", post:"Population", target:str, name:str=None, copied=False, net_id=0):
        """
        """
        # Instantiate the projection
        SpecificProjection.__init__(self, pre, post, target, None, name, copied, net_id)

        # Check populations
        if not self.pre.neuron_type.type == 'rate':
            Messages._error('The pre-synaptic population of a CurrentInjection must be rate-coded.')

        if not self.post.neuron_type.type == 'spike':
            Messages._error('The post-synaptic population of a CurrentInjection must be spiking.')

        if not self.post.size == self.pre.size:
            Messages._error('CurrentInjection: The pre- and post-synaptic populations must have the same size.')

        if _check_paradigm("cuda", self.net_id) and (isinstance(pre, PopulationView) or isinstance(post, PopulationView)):
            Messages._error("CurrentInjection on GPUs is not allowed for PopulationViews")

        # Prevent automatic split of matrices
        self._no_split_matrix = True

    def _copy(self, pre, post, net_id=None):
        "Returns a copy of the population when creating networks. Internal use only."
        return CurrentInjection(pre=pre, post=post, target=self.target, name=self.name, copied=True, net_id = self.net_id if net_id is None else net_id)

    def _generate_st(self):
        # Generate the code
        self._specific_template['psp_code'] = """
        if (pop%(id_post)s->_active) {
            for (int i=0; i<post_rank.size(); i++) {
                pop%(id_post)s->g_%(target)s[post_rank[i]] += pop%(id_pre)s->r[pre_rank[i][0]];
            }
        } // active
""" % { 'id_pre': self.pre.id, 'id_post': self.post.id, 'target': self.target}

    def _generate_omp(self):
        # Generate the code
        self._specific_template['psp_code'] = """
        if (pop%(id_post)s->_active) {
            #pragma omp for
            for (int i=0; i<post_rank.size(); i++) {
                pop%(id_post)s->g_%(target)s[post_rank[i]] += pop%(id_pre)s->r[pre_rank[i][0]];
            }
        } // active
""" % { 'id_pre': self.pre.id, 'id_post': self.post.id, 'target': self.target}

    def _generate_cuda(self):
        """
        Generate the CUDA code.

        For a first implementation we take a rather simple approach:

        * We use only one block for the kernel and each thread computes
        one synapse/post-neuron entry (which is equal as we use one2one).

        * We use only the pre-synaptic firing rate, no other variables.

        * We ignore the synaptic weight.
        """
        ids = {
            'id_proj': self.id,
            'id_post': self.post.id,
            'id_pre': self.pre.id,
            'target': self.target,
            'float_prec': ConfigManager().get('precision', self.net_id)
        }

        self._specific_template['psp_body'] = """
__global__ void cu_proj%(id_proj)s_psp(int post_size, %(float_prec)s *pre_r, %(float_prec)s *g_%(target)s) {
    int n = threadIdx.x;

    while (n < post_size) {
        g_%(target)s[n] += pre_r[n];

        n += blockDim.x;
    }
}
""" % ids
        self._specific_template['psp_invoke'] = """
void proj%(id_proj)s_psp(RunConfig cfg, int post_size, %(float_prec)s *pre_r, %(float_prec)s *g_%(target)s) {
    cu_proj%(id_proj)s_psp<<< cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream >>>(post_size, pre_r, g_%(target)s);
}
""" % ids
        self._specific_template['psp_header'] = """void proj%(id_proj)s_psp(RunConfig cfg, int post_size, %(float_prec)s *pre_r, %(float_prec)s *g_%(target)s);""" % ids
        self._specific_template['psp_call'] = """
    proj%(id_proj)s_psp(
        RunConfig(1, 192, 0, proj%(id_proj)s->stream),
        pop%(id_post)s->size,
        pop%(id_pre)s->gpu_r,
        pop%(id_post)s->gpu_g_%(target)s
    );
""" % ids

    def connect_current(self):
        return self.one_to_one(weights=1.0)


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

# Code which should be added prior to kernels
additional_global_functions = ""

launch_config = {
    'init': """
        _threads_per_block = 32;
        auto tmp_blocks = static_cast<unsigned int>(ceil(static_cast<double>(nb_dendrites())/static_cast<double>(_threads_per_block)));
        _nb_blocks = std::min<unsigned int>(tmp_blocks, 65535);

    #ifdef _DEBUG
        std::cout << "Initial kernel configuration: " << _nb_blocks << ", " << _threads_per_block << std::endl;
    #endif
""",
    'update': """
        if (nb_blocks != -1) {
            _nb_blocks = static_cast<unsigned int>(nb_blocks);
            _threads_per_block = threads_per_block;
        }else{
            _threads_per_block = threads_per_block;
            auto tmp_blocks = static_cast<unsigned int>(ceil(static_cast<double>(nb_dendrites())/static_cast<double>(_threads_per_block)));
            _nb_blocks = std::min<unsigned int>(tmp_blocks, 65535);
        }

    #ifdef _DEBUG
        std::cout << "Updated kernel configuration: " << _nb_blocks << ", " << _threads_per_block << std::endl;
    #endif
"""
}

attribute_decl = {
    'local': """
    // Local %(attr_type)s %(name)s
    std::vector< %(type)s > %(name)s;
    %(type)s* gpu_%(name)s;
    bool %(name)s_host_to_device;
    long int %(name)s_device_to_host;
""",
    'semiglobal': """
    // Semiglobal %(attr_type)s %(name)s
    std::vector< %(type)s >  %(name)s ;
    %(type)s* gpu_%(name)s;
    bool %(name)s_host_to_device;
    long int %(name)s_device_to_host;
""",
    'global': {
        'parameter':
    """
    // Global parameter %(name)s
    %(type)s %(name)s;
""",
        'variable': """
    // Global variable %(name)s
    %(type)s %(name)s;
    %(type)s* gpu_%(name)s;
    bool %(name)s_host_to_device;
    long int %(name)s_device_to_host;
"""
    }
}

attribute_cpp_init = {
    'local': """
        // Local %(attr_type)s %(name)s
        %(name)s = init_matrix_variable<%(type)s>(%(init)s);
        gpu_%(name)s = init_matrix_variable_gpu<%(type)s>(%(name)s);
        %(name)s_host_to_device = false;
        %(name)s_device_to_host = t;
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        %(name)s = init_vector_variable<%(type)s>(%(init)s);
        gpu_%(name)s = init_vector_variable_gpu<%(type)s>(%(name)s);
        %(name)s_host_to_device = false;
        %(name)s_device_to_host = t;
""",
    'global': { 
        'parameter': """
        // Global parameter %(name)s
        %(name)s = 0.0;
""",
        'variable': """
        // Global variable %(name)s
        %(name)s = static_cast<%(type)s>(%(init)s);
        cudaMalloc((void**)&gpu_%(name)s, sizeof(%(type)s));
        %(name)s_host_to_device = true;
        %(name)s_device_to_host = t;
"""
    }
}

attribute_cpp_size = {
    'local': """
        // Local %(attr_type)s %(name)s
        size_in_bytes += sizeof(bool);
        size_in_bytes += sizeof(%(ctype)s*);
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        size_in_bytes += sizeof(bool);
        size_in_bytes += sizeof(%(ctype)s*);
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'global': """
        // Global
        size_in_bytes += sizeof(bool);
        size_in_bytes += sizeof(%(ctype)s*);
        size_in_bytes += sizeof(%(ctype)s);
"""
}

attribute_cpp_delete = {
    'local': """
        // %(name)s - host
        %(name)s.clear();
        %(name)s.shrink_to_fit();

        // %(name)s - device
        cudaFree(gpu_%(name)s);
""",
    'semiglobal': """
        // %(name)s - host
        %(name)s.clear();
        %(name)s.shrink_to_fit();

        // %(name)s - device
        cudaFree(gpu_%(name)s);
""",
    'global': ""
}

attribute_host_to_device = {
    'local': """
        // %(name)s: local
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_%(name)s, %(name)s.data(), %(name)s.size() * sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
""",
    'semiglobal': """
        // %(name)s: semiglobal
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_%(name)s, %(name)s.data(), %(name)s.size() * sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
""",
    'global': """
        // %(name)s: global
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_%(name)s, &%(name)s, sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
"""
}

attribute_device_to_host = {
    'local': """
        // %(name)s: local
        if ( %(name)s_device_to_host < t ) {
        #ifdef _DEBUG
            std::cout << "DtoH: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( %(name)s.data(), gpu_%(name)s, %(name)s.size() * sizeof( %(type)s ), cudaMemcpyDeviceToHost);
        #ifdef _DEBUG
            cudaError_t err_%(name)s = cudaGetLastError();
            if ( err_%(name)s != cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err_%(name)s) << std::endl;
        #endif
            %(name)s_device_to_host = t;
        }
""",
    'semiglobal': "",
    'global': ""
}

#
#   Synaptic delays
#
delay = {
    # A single value for all synapses
    'uniform': {
        'declare': """
    // Uniform delay
    int delay;

    int get_delay() { return delay; }
    int get_dendrite_delay(int idx) { return delay; }
    void set_delay(int delay) { this->delay = delay; }
""",
        'init': """
    delay = delays[0][0];
""",
    },
    # An individual value for each synapse
    'nonuniform_rate_coded': None,
    # An individual value for each synapse
    'nonuniform_spiking': None
}

#
# Implement the continuous transmission for rate-code synapses.
#
rate_psp_kernel = {
    'device_kernel': {
        'sum':"""
__global__ void cu_proj%(id_proj)s_psp(%(conn_args)s%(add_args)s, %(float_prec)s* %(target_arg)s ) {
    int i = blockIdx.x * blockDim.x + threadIdx.x;

    while( i < post_size ) {
        
        %(float_prec)s localSum = %(thread_init)s;

        for(int j =0; j < rl[i]; j++)
            localSum += %(psp)s

        %(target_arg)s%(post_index)s += localSum;

        i += gridDim.x * blockDim.x;
    }
}
"""
    },
    'invoke_kernel': """
void proj%(id_proj)s_psp(RunConfig cfg, %(conn_args)s%(add_args)s, %(float_prec)s* %(target_arg)s) {
    cu_proj%(id_proj)s_psp<<<cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream>>>(
        /* ranks and offsets */
        %(conn_args_call)s
        /* computation data */
        %(add_args_call)s
        /* result */
        %(target_arg_call)s
    );
}
""",
    'kernel_decl': """void proj%(id_proj)s_psp(RunConfig cfg, %(conn_args)s%(add_args)s, %(float_prec)s* %(target_arg)s );
""",
    'host_call': """
    // proj%(id_proj)s: pop%(id_pre)s -> pop%(id_post)s
    if ( pop%(id_post)s->_active && proj%(id_proj)s->_transmission ) {

        proj%(id_proj)s_psp(
            RunConfig(proj%(id_proj)s->_nb_blocks, proj%(id_proj)s->_threads_per_block, 0, proj%(id_proj)s->stream),
            /* ranks and offsets */
            %(conn_args)s
            /* computation data */
            %(add_args)s
            /* result */
            %(target_arg)s
        );

    #ifdef _DEBUG
        auto err = cudaGetLastError();
        if ( err != cudaSuccess ) {
            std::cout << "cu_proj%(id_proj)s_psp: " << cudaGetErrorString(err) << std::endl;
        }
    #endif
    }
""",
    'thread_init': {
        'float': {
            'sum': "0.0f",
            'min': "FLT_MAX",
            'max': "FLT_MIN",
            'mean': "0.0f"
        },
        'double': {
            'sum': "0.0",
            'min': "DBL_MAX",
            'max': "DBL_MIN",
            'mean': "0.0"
        }
    }    
}


# Update of global synaptic equations, consist of body (annarchyDevice.cu),
# header and call semantic (take place in ANNarchyHost.cu)
global_synapse_update = {
    'device_kernel': """
// gpu device kernel for projection %(id_proj)s
__global__ void cuProj%(id_proj)s_global_step(
    /* default params */
    const long int t, %(float_prec)s dt
    /* additional params */
    %(kernel_args)s,
    /* plasticity enabled */
    bool plasticity
) {
%(pre_loop)s
%(global_eqs)s
}
""",
    'invoke_kernel': """void proj%(id_proj)s_global_step(RunConfig cfg, const long int t, %(float_prec)s dt %(kernel_args)s, bool plasticity) {
    cuProj%(id_proj)s_global_step<<<cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream>>>(
        /* default args*/
        t, dt
        /* kernel args */
        %(kernel_args_call)s
        /* synaptic plasticity */
        , plasticity
    );
}
""",
    'kernel_decl': """void proj%(id_proj)s_global_step(RunConfig cfg, const long int t, %(float_prec)s dt %(kernel_args)s, bool plasticity);
""",
    'host_call': """
        // global update
        proj%(id_proj)s_global_step(
            /* kernel configuration */
            RunConfig(1, 1, 0, proj%(id_proj)s->stream),
            /* default args*/
            t, _dt
            %(kernel_args_call)s
            /* synaptic plasticity */
            , proj%(id_proj)s->_plasticity
        );

    #ifdef _DEBUG
        cudaDeviceSynchronize();
        err = cudaGetLastError();
        if ( global_step != cudaSuccess) {
            std::cout << "proj%(id_proj)s_step: " << cudaGetErrorString( err ) << std::endl;
        }
    #endif
"""
}

# Update of semiglobal synaptic equations, consist of body (annarchyDevice.cu),
# header and call semantic (take place in ANNarchyHost.cu)
semiglobal_synapse_update = {
    'device_kernel': """
// gpu device kernel for projection %(id_proj)s
__global__ void cuProj%(id_proj)s_semiglobal_step(
    %(idx_type)s post_size, const %(idx_type)s* __restrict__ rank_post,
    /* default params */
    const long int t, %(float_prec)s dt
    /* additional params */
    %(kernel_args)s,
    /* plasticity enabled */
    bool plasticity 
) {
    int i = threadIdx.x + blockIdx.x*blockDim.x;
    

%(pre_loop)s
    while ( i < post_size ) {
        %(idx_type)s rk_post = rank_post[i];

%(semiglobal_eqs)s

        i += gridDim.x * blockDim.x;
    }
}
""",
    'invoke_kernel': """
void proj%(id_proj)s_semiglobal_step(RunConfig cfg, %(idx_type)s post_size, const %(idx_type)s* __restrict__ rank_post, const long int t, %(float_prec)s dt %(kernel_args)s, bool plasticity) {
    cuProj%(id_proj)s_semiglobal_step<<< cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream >>>(
        /* connectivity */
        post_size, rank_post,
        /* default args*/
        t, dt
        /* kernel args */
        %(kernel_args_call)s
        /* synaptic plasticity */
        , plasticity
    );

}
""",
    'kernel_decl': """void proj%(id_proj)s_semiglobal_step(RunConfig cfg, %(idx_type)s post_size, const %(idx_type)s* __restrict__ rank_post, const long int t, %(float_prec)s dt %(kernel_args)s, bool plasticity);
""",
    'host_call': """
        // semiglobal update
        proj%(id_proj)s_semiglobal_step(
            RunConfig(proj%(id_proj)s->_nb_blocks, proj%(id_proj)s->_threads_per_block, 0, proj%(id_proj)s->stream),
            /* connectivity */
            proj%(id_proj)s->nb_dendrites(), proj%(id_proj)s->gpu_post_ranks_,
            /* default args*/
            t, dt
            /* kernel args */
            %(kernel_args_call)s
            /* synaptic plasticity */
            , proj%(id_proj)s->_plasticity
        );

    #ifdef _DEBUG
        cudaDeviceSynchronize();
        err = cudaGetLastError();
        if ( err != cudaSuccess) {
            std::cout << "proj%(id_proj)s_semiglobal_step: " << cudaGetErrorString( err ) << std::endl;
        }
    #endif
"""
}

# Update of local synaptic equations, consist of body (annarchyDevice.cu),
# header and call semantic (take place in ANNarchyHost.cu)
local_synapse_update = {
    'device_kernel': """
// gpu device kernel for projection %(id_proj)s
__global__ void cuProj%(id_proj)s_local_step(
    /* connectivity */
    %(conn_args)s,
    /* default params */
    const long int t, %(float_prec)s dt
    /* additional params */
    %(kernel_args)s,
    /* plasticity enabled */
    bool plasticity
) {
    int i = blockIdx.x * blockDim.x + threadIdx.x;
%(pre_loop)s

    // Updating local variables of projection %(id_proj)s
    while ( i < post_size )
    {
        for ( int j = 0; j < rl[i]; j++ ) {
%(local_eqs)s
        }

        i += blockDim.x * gridDim.x;
    }
}
""",
    'invoke_kernel': """
void proj%(id_proj)s_local_step(RunConfig cfg, %(conn_args)s, const long int t, %(float_prec)s dt %(kernel_args)s, bool plasticity) {
    cuProj%(id_proj)s_local_step<<<cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream>>>(
        /* connectivity */
        %(conn_args_call)s
        /* default args*/
        , t, dt
        /* kernel args */
        %(kernel_args_call)s
        /* synaptic plasticity */
        , plasticity
    );
}
""",
    'kernel_decl': """void proj%(id_proj)s_local_step(RunConfig cfg, %(conn_args)s, const long int t, %(float_prec)s dt %(kernel_args)s, bool plasticity);
""",
    'host_call': """
        // local update
        proj%(id_proj)s_local_step(
            /* kernel configuration */
            RunConfig(1, 32, 0, proj%(id_proj)s->stream),
            /* connectivity */
            %(conn_args_call)s
            /* default args*/
            , t, _dt
            /* kernel args */
            %(kernel_args_call)s
            /* synaptic plasticity */
            , proj%(id_proj)s->_plasticity
        );

    #ifdef _DEBUG
        cudaDeviceSynchronize();
        err = cudaGetLastError();
        if ( err != cudaSuccess) {
            std::cout << "proj%(id_proj)s_step: " << cudaGetErrorString( err ) << std::endl;
        }
    #endif
"""
}

# call semantic for global, semiglobal and local kernel
synapse_update_call = """
    // proj%(id_proj)s: pop%(pre)s -> pop%(post)s
    if ( proj%(id_proj)s->_transmission && proj%(id_proj)s->_update && proj%(id_proj)s->_plasticity && ( (t - proj%(id_proj)s->_update_offset)%%proj%(id_proj)s->_update_period == 0L)) {
        %(float_prec)s _dt = dt * proj%(id_proj)s->_update_period;
#ifdef _DEBUG
    cudaError_t err;
#endif
%(global_call)s
int nb_blocks;
%(semiglobal_call)s
%(local_call)s
    }
"""

conn_templates = {
    # connectivity representation
    'conn_header': "const %(idx_type)s post_size, const %(idx_type)s* __restrict__ rank_post, const %(idx_type)s* __restrict__ rank_pre, const %(idx_type)s* __restrict__ rl",
    'conn_call': "proj%(id_proj)s->nb_dendrites(), proj%(id_proj)s->gpu_post_ranks_, proj%(id_proj)s->gpu_col_idx_, proj%(id_proj)s->gpu_rl_",
    'conn_kernel': "post_size, rank_post, rank_pre, rl",

    # launch config
    'launch_config': launch_config,
 
    # accessors
    'attribute_decl': attribute_decl,
    'attribute_cpp_init': attribute_cpp_init,
    'attribute_cpp_size': attribute_cpp_size,
    'attribute_cpp_delete': attribute_cpp_delete,
    'host_to_device': attribute_host_to_device,
    'device_to_host': attribute_device_to_host,
    'delay': delay,

    # operations
    'rate_psp': rate_psp_kernel,
    'synapse_update': {
        'global': global_synapse_update,
        'semiglobal': semiglobal_synapse_update,
        'local': local_synapse_update,
        'call': synapse_update_call
    }
}

conn_ids = {
    'local_index': "[j*post_size+i]",
    'semiglobal_index': '[i]',
    'global_index': '[0]',
    'pre_index': '[rk_pre]',
    'post_index': '[rk_post]',
    'pre_prefix': 'pre_',
    'post_prefix': 'post_'
}


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

omp_header_template = """#pragma once

#include <string>
#include <vector>
#include <algorithm>
#include <map>
#include <deque>
#include <queue>
#include <iostream>
#include <sstream>
#include <fstream>
#include <cstdlib>
#include <stdlib.h>
#include <string.h>
#include <cmath>
#include <random>
#include <cassert>
// only included if compiled with -fopenmp
#ifdef _OPENMP
    #include <omp.h>
#endif

// Intrinsic operations (Intel/AMD)
#ifdef __x86_64__
    #include <immintrin.h>
#endif

/*
 * Built-in functions
 *
 */
%(built_in)s

/*
 * Custom constants
 *
 */
%(custom_constant)s

/*
 * Custom functions
 *
 */
%(custom_func)s

/*
 * Structures for the populations
 *
 */
%(pop_struct)s
/*
 * Structures for the projections
 *
 */
%(proj_struct)s

/*
 * Declaration of the populations
 *
 */
%(pop_ptr)s

/*
 * Declaration of the projections
 *
 */
%(proj_ptr)s

/*
 * Recorders
 *
 */
#include "Monitor.hpp"

extern std::vector<Monitor*> recorders;
int addRecorder(Monitor* recorder);
Monitor* getRecorder(int id);
void removeRecorder(Monitor* recorder);

/*
 * Simulation methods
 *
 */
void run(const int nbSteps);
int run_until(const int steps, std::vector<int> populations, bool or_and);
void step();

/*
 *  Initialization
 */
void initialize(const %(float_prec)s dt_) ;

/*
 *  Life-time management
 */
void create_cpp_instances();
void destroy_cpp_instances();

/*
 * Time export
 *
 */
long int getTime();
void setTime(const long int t_);
%(float_prec)s getDt();
void setDt(const %(float_prec)s dt_);

/*
 * Number of threads
 *
 */
void setNumberThreads(int threads, std::vector<int> core_list);

/*
 * Seed for the RNG
 *
*/
void setSeed(long int seed, int num_sources, bool use_seed_seq);

"""

st_body_template = """
#include "ANNarchy.hpp"

%(prof_include)s

/*
 * Internal data
 *
 */
%(float_prec)s dt;
long int t;
std::vector<std::mt19937> rng;

// Custom constants
%(custom_constant)s

// Populations
%(pop_ptr)s

// Projections
%(proj_ptr)s

// Global operations
%(glops_def)s

/*
 * Recorders
 */
std::vector<Monitor*> recorders;
int addRecorder(Monitor* recorder){
    int found = -1;

    for (unsigned int i=0; i<recorders.size(); i++) {
        if (recorders[i] == nullptr) {
            found = i;
            break;
        }
    }

    if (found != -1) {
        // fill a previously cleared slot
        recorders[found] = recorder;
        return found;
    } else {
        recorders.push_back(recorder);
        return recorders.size() - 1;
    }
}

Monitor* getRecorder(int id) {
    if (id < recorders.size())
        return recorders[id];
    else
        return nullptr;
}

void removeRecorder(Monitor* recorder){
    for (unsigned int i=0; i<recorders.size(); i++){
        if (recorders[i] == recorder) {
            // mark the slot as free
            recorders[i] = nullptr;
            break;
        }
    }
}

/*
 *  Simulation methods
 */
// Simulate a single step
void singleStep()
{
#ifdef _TRACE_SIMULATION_STEPS
    std::cout << "-- evaluate step " << t << " (" << t * dt << " ms) --" << std::endl;
#endif
%(prof_step_pre)s

    ////////////////////////////////
    // Presynaptic events
    ////////////////////////////////
%(prof_proj_psp_pre)s
%(reset_sums)s
#ifdef _TRACE_SIMULATION_STEPS
    std::cout << "Update psp/conductances ..." << std::endl;
#endif
%(compute_sums)s
%(prof_proj_psp_post)s

    ////////////////////////////////
    // Recording target variables
    ////////////////////////////////
#ifdef _TRACE_SIMULATION_STEPS
    std::cout << "Record psp/conductances ..." << std::endl;
#endif
    for (unsigned int i=0; i < recorders.size(); i++) {
        if (recorders[i])
            recorders[i]->record_targets();
    }

    ////////////////////////////////
    // Update random distributions
    ////////////////////////////////
#ifdef _TRACE_SIMULATION_STEPS
    std::cout << "Draw required random numbers ..." << std::endl;
#endif
%(prof_rng_pre)s
%(random_dist_update)s
%(prof_rng_post)s

    ////////////////////////////////
    // Update neural variables
    ////////////////////////////////
#ifdef _TRACE_SIMULATION_STEPS
    std::cout << "Evaluate neural ODEs ..." << std::endl;
#endif
%(prof_neur_step_pre)s
%(update_neuron)s
%(prof_neur_step_post)s

    ////////////////////////////////
    // Delay outputs
    ////////////////////////////////
#ifdef _TRACE_SIMULATION_STEPS
    std::cout << "Update delay queues ..." << std::endl;
#endif
%(delay_code)s

    ////////////////////////////////
    // Global operations (min/max/mean)
    ////////////////////////////////
#ifdef _TRACE_SIMULATION_STEPS
    std::cout << "Update global operations ..." << std::endl;
#endif
%(prof_global_ops_pre)s
%(update_globalops)s
%(prof_global_ops_post)s

    ////////////////////////////////
    // Update synaptic variables
    ////////////////////////////////
#ifdef _TRACE_SIMULATION_STEPS
    std::cout << "Evaluate synaptic ODEs ..." << std::endl;
#endif
%(prof_proj_step_pre)s
%(update_synapse)s
%(prof_proj_step_post)s

    ////////////////////////////////
    // Postsynaptic events
    ////////////////////////////////
%(prof_proj_post_event_pre)s
%(post_event)s
%(prof_proj_post_event_post)s

    ////////////////////////////////
    // Structural plasticity
    ////////////////////////////////
%(structural_plasticity)s

    ////////////////////////////////
    // Recording neural / synaptic variables
    ////////////////////////////////
%(prof_record_pre)s
    for (unsigned int i=0; i < recorders.size(); i++){
        if (recorders[i])
            recorders[i]->record();
    }
%(prof_record_post)s

    ////////////////////////////////
    // Increase internal time
    ////////////////////////////////
    t++;

%(prof_step_post)s

#ifdef _TRACE_SIMULATION_STEPS
    std::cout << "-- simulation step " << t << " completed --" << std::endl;
#endif
}

// Simulate the network for the given number of steps,
// called from python
void run(const int nbSteps) {
#ifdef _TRACE_SIMULATION_STEPS
    std::cout << "Perform simulation for " << nbSteps << " steps." << std::endl;
#endif
%(prof_run_pre)s
    // apply changes implied by structural plasticity (spike only)
%(sp_spike_backward_view_update)s

    // perform the simulation
    for(int i=0; i<nbSteps; i++) {
        singleStep();
    }
%(prof_run_post)s
}

// Simulate the network for a single steps,
// called from python
void step() {
%(prof_run_pre)s
    // apply changes implied by structural plasticity (spike only)
%(sp_spike_backward_view_update)s

    // perform a single step (size dt)
    singleStep();
%(prof_run_post)s
}

int run_until(const int steps, std::vector<int> populations, bool or_and)
{
    // apply changes implied by structural plasticity (spike only)
%(sp_spike_backward_view_update)s

    // perform the simulation until the condition is satisfied
%(run_until)s
}

/*
 *  Initialization methods
 */
// Initialize the internal data and the random numbers generator
void initialize(const %(float_prec)s _dt) {
%(initialize)s
}

// Change the seed of the RNG
void setSeed(const long int seed, const int num_sources, const bool use_seed_seq) {
#ifdef _DEBUG
    std::cout << "ANNarchyCore::setSeed(): " << seed << ", " << num_sources << ", " << std::string((use_seed_seq) ? "true" : "false") << std::endl;
#endif
    // sanity check
    if (num_sources > 1)
        std::cerr << "WARNING - ANNarchyCore::setSeed(): num_sources should be 1 for single thread code." << std::endl;

    rng.clear();

    rng.push_back(std::mt19937(seed));

    rng.shrink_to_fit();
}

/*
 *  Life-time management
 */
void create_cpp_instances() {
#if defined(_TRACE_INIT) || defined(_DEBUG)
    std::cout << "Instantiate C++ objects ..." << std::endl;
#endif
}

void destroy_cpp_instances() {
#if defined(_TRACE_INIT) || defined(_DEBUG)
    std::cout << "Destroy C++ objects ..." << std::endl;
#endif
}

/*
 * Access to time and dt
 */
long int getTime() {return t;}
void setTime(const long int t_) { t=t_;}
%(float_prec)s getDt() { return dt;}
void setDt(const %(float_prec)s dt_) { dt=dt_;}

/*
 * Number of threads
 *
*/
void setNumberThreads(const int threads, const std::vector<int> core_list)
{
    if (threads > 1) {
        std::cerr << "WARNING: a call of setNumberThreads() is without effect on single thread simulation code." << std::endl;
    }

    if (core_list.size()>1) {
        std::cerr << "The provided core list is ambiguous and therefore ignored." << std::endl;
        return;
    }

#ifdef __linux__
    // set a cpu mask to prevent moving of threads
    cpu_set_t mask;

    // no CPUs selected
    CPU_ZERO(&mask);

    // no proc_bind
    for(auto it = core_list.begin(); it != core_list.end(); it++)
        CPU_SET(*it, &mask);
    const int set_result = sched_setaffinity(0, sizeof(cpu_set_t), &mask);
#else
    if (!core_list.empty()) {
        std::cout << "WARNING: manipulation of CPU masks is only available for linux systems." << std::endl;
    }
#endif
}
"""

omp_body_template = """
#include "ANNarchy.hpp"

#ifdef __linux__
#include <sched.h>
#endif

%(prof_include)s

/*
 * Internal data
 *
 */
%(float_prec)s dt;
long int t;
std::vector<std::mt19937> rng;

// number openMP threads
int global_num_threads = -1;

// Custom constants
%(custom_constant)s

// Populations
%(pop_ptr)s

// Projections
%(proj_ptr)s

// Global operations
%(glops_def)s

/*
 * Recorders
 */
std::vector<Monitor*> recorders;
int addRecorder(Monitor* recorder){
    int found = -1;

    for (unsigned int i=0; i<recorders.size(); i++) {
        if (recorders[i] == nullptr) {
            found = i;
            break;
        }
    }

    if (found != -1) {
        // fill a previously cleared slot
        recorders[found] = recorder;
        return found;
    } else {
        recorders.push_back(recorder);
        return recorders.size() - 1;
    }
}

Monitor* getRecorder(int id) {
    if (id < recorders.size())
        return recorders[id];
    else
        return nullptr;
}

void removeRecorder(Monitor* recorder){
    for (unsigned int i=0; i<recorders.size(); i++){
        if (recorders[i] == recorder) {
            // mark position as free
            recorders[i] = nullptr;
            break;
        }
    }
}

/*
 *  Simulation methods
 */
// Step method. Generated by ANNarchy.
void singleStep(const int tid, const int nt)
{
#ifdef _TRACE_SIMULATION_STEPS
    #pragma omp single
    {
        std::cout << "-- evaluate step " << t << " (" << t * dt << " ms) --" << std::endl;
        std::cout << std::flush;
    }
#endif
%(prof_step_pre)s

    ////////////////////////////////
    // Presynaptic events
    ////////////////////////////////
%(prof_proj_psp_pre)s
%(reset_sums)s
#ifdef _TRACE_SIMULATION_STEPS
    #pragma omp single
    {
        std::cout << "Update psp/conductances ..." << std::endl;
        std::cout << std::flush;
    }
#endif
%(compute_sums)s

    #pragma omp barrier
%(prof_proj_psp_post)s

    ////////////////////////////////
    // Recording target variables
    ////////////////////////////////
#ifdef _TRACE_SIMULATION_STEPS
    #pragma omp single
    {
        std::cout << "Record psp/conductances ..." << std::endl;
        std::cout << std::flush;
    }
#endif
    for (unsigned int i=tid; i < recorders.size(); i += nt) {
        if (recorders[i])
            recorders[i]->record_targets();
    }

    #pragma omp barrier

    ////////////////////////////////
    // Update random distributions
    ////////////////////////////////
#ifdef _TRACE_SIMULATION_STEPS
    #pragma omp single
    {
        std::cout << "Draw required random numbers ..." << std::endl;
        std::cout << std::flush;
    }
#endif
%(prof_rng_pre)s
%(random_dist_update)s
%(prof_rng_post)s

    ////////////////////////////////
    // Update neural variables
    ////////////////////////////////
#ifdef _TRACE_SIMULATION_STEPS
    #pragma omp single
    {
        std::cout << "Evaluate neural ODEs ..." << std::endl;
        std::cout << std::flush;
    }
#endif
%(prof_neur_step_pre)s
%(update_neuron)s
%(prof_neur_step_post)s

    #pragma omp barrier

    ////////////////////////////////
    // Delay outputs
    ////////////////////////////////
#ifdef _TRACE_SIMULATION_STEPS
    #pragma omp single
    {
        std::cout << "Update delay queues ..." << std::endl;
        std::cout << std::flush;
    }
#endif
%(delay_code)s

    ////////////////////////////////
    // Global operations (min/max/mean)
    ////////////////////////////////
#ifdef _TRACE_SIMULATION_STEPS
    #pragma omp single
    {
        std::cout << "Update global operations ..." << std::endl;
        std::cout << std::flush;
    }
#endif
%(prof_global_ops_pre)s
%(update_globalops)s
%(prof_global_ops_post)s

    ////////////////////////////////
    // Update synaptic variables
    ////////////////////////////////
#ifdef _TRACE_SIMULATION_STEPS
    #pragma omp single
    {
        std::cout << "Evaluate synaptic ODEs ..." << std::endl;
        std::cout << std::flush;
    }
#endif
%(prof_proj_step_pre)s
%(update_synapse)s
%(prof_proj_step_post)s

    #pragma omp barrier

    ////////////////////////////////
    // Postsynaptic events
    ////////////////////////////////
%(post_event)s

    ////////////////////////////////
    // Structural plasticity
    ////////////////////////////////
%(structural_plasticity)s

    ////////////////////////////////
    // Recording neural / synaptic variables
    ////////////////////////////////
    #pragma omp barrier
#ifdef _TRACE_SIMULATION_STEPS
    #pragma omp single
    {
        std::cout << "Recording state variables ..." << std::endl;
        std::cout << std::flush;
    }
#endif

%(prof_record_pre)s
    for (unsigned int i=tid; i < recorders.size(); i += nt){
        if (recorders[i])
            recorders[i]->record();
    }
%(prof_record_post)s

    #pragma omp barrier

    ////////////////////////////////
    // Increase internal time
    ////////////////////////////////
    #pragma omp single
    {
        t++;
    } // implicit barrier

%(prof_step_post)s

#ifdef _TRACE_SIMULATION_STEPS
    #pragma omp single
    {
        std::cout << "-- simulation step " << t << " completed --" << std::endl;
        std::cout << std::flush;
    }
#endif
}

// Simulate the network for the given number of steps,
// called from python
void run(const int nbSteps) {
#ifdef _TRACE_SIMULATION_STEPS
    std::cout << "Perform simulation for " << nbSteps << " steps." << std::endl;
#endif
%(prof_run_pre)s
    // apply changes implied by structural plasticity (spike only)
%(sp_spike_backward_view_update)s

    // perform the simulation
    #pragma omp parallel num_threads(global_num_threads)
    {
        int tid = omp_get_thread_num();

        for (int i=0; i<nbSteps; i++) {
            singleStep(tid, global_num_threads);
        }
    }

%(prof_run_post)s
}

// Simulate the network for a single steps,
// called from python
void step() {
%(prof_run_pre)s
    // apply changes implied by structural plasticity (spike only)
%(sp_spike_backward_view_update)s

    // perform a single step (size dt)
    #pragma omp parallel num_threads(global_num_threads)
    {
        int tid = omp_get_thread_num();

        singleStep(tid, global_num_threads);
    }
%(prof_run_post)s
}

int run_until(const int steps, std::vector<int> populations, bool or_and)
{
    // apply changes implied by structural plasticity (spike only)
%(sp_spike_backward_view_update)s

    // perform the simulation until the condition is satisfied
%(run_until)s
}

/*
 *  Initialization methods
 */
// Initialize the internal data and the random numbers generator
void initialize(const %(float_prec)s _dt) {
%(initialize)s
}

// Change the seed of the RNG
void setSeed(const long int seed, const int num_sources, const bool use_seed_seq){
#ifdef _DEBUG
    std::cout << "ANNarchyCore::setSeed(): " << seed << ", " << num_sources << ", " << std::string((use_seed_seq) ? "true" : "false") << std::endl;
#endif
    // sanity check
    if (num_sources > 1) {
        if (num_sources != global_num_threads) {
            std::cerr << "WARNING - ANNarchyCore::setSeed(): num_sources should be equal to the number of used OpenMP threads." << std::endl;
            std::cerr << "    got: " << num_sources << " and expected: " << global_num_threads << std::endl;
        }
    }

    rng.clear();

    if (num_sources == 1) {
        rng.push_back(std::mt19937(seed));
    } else {
        if (use_seed_seq) {
            std::seed_seq seq{seed};
            std::vector<std::uint32_t> seeds(num_sources);
            seq.generate(seeds.begin(), seeds.end());

            for (auto it = seeds.begin(); it != seeds.end(); it++) {
                rng.push_back(std::mt19937(*it));
            }
        } else {
            // Using seed initialization of M.E. O'Neill (randutils)
            std::vector<std::uint32_t> seeds(num_sources);
            randutils::auto_seed_256 seeder;
            seeder.generate(seeds.begin(), seeds.end());

            for (auto it = seeds.begin(); it != seeds.end(); it++) {
                rng.push_back(std::mt19937(*it));
            }
        }
    }

    rng.shrink_to_fit();
}

/*
 *  Life-time management
 */
void create_cpp_instances() {
#ifdef _DEBUG
    std::cout << "Instantiate C++ objects ..." << std::endl;
#endif
}

void destroy_cpp_instances() {
#ifdef _DEBUG
    std::cout << "Destroy C++ objects ..." << std::endl;
#endif
}

/*
 * Access to time and dt
 */
long int getTime() {return t;}
void setTime(const long int t_) { t=t_;}
%(float_prec)s getDt() { return dt;}
void setDt(const %(float_prec)s dt_) { dt=dt_;}

/*
 * Number of threads
 *
 */
void setNumberThreads(const int threads, const std::vector<int> core_list)
{
#ifdef _DEBUG
    std::cout << "Set new number of threads:" << threads << std::endl;
    if (!core_list.empty()) {
        std::cout << "Use thread placement: [";
        for (auto it = core_list.begin(); it != core_list.end(); it++) std::cout << *it << " ";
        std::cout << "]";
    }
#endif

    // set worker set size
    global_num_threads = threads;

#ifdef __linux__
    // set a cpu mask to prevent moving of threads
    cpu_set_t mask;

    // no CPUs selected
    CPU_ZERO(&mask);

    // no proc_bind
    for(auto it = core_list.begin(); it != core_list.end(); it++)
        CPU_SET(*it, &mask);
    const int set_result = sched_setaffinity(0, sizeof(cpu_set_t), &mask);
#else
    if (!core_list.empty()) {
        std::cout << "WARNING: manipulation of CPU masks is only available for linux systems." << std::endl;
    }
#endif
}
"""

omp_run_until_template = {
    'default':
"""
    run(steps);
    return steps;
""",
    'body':
"""
    bool stop = false;
    bool cond_activated = false;
    int nb = 0;
    for(int n = 0; n < steps; n++)
    {
        step();
        nb++;
        stop = or_and;

%(run_until)s

        // HD: stop will be automatically true, if no populations are checked.
        if(stop && (populations.size() > 0))
            break;
    }
    return nb;

""",
    'single_pop': """
        cond_activated = std::find(populations.begin(), populations.end(), %(id)s) != populations.end();
        if (cond_activated)
            if(or_and)
                stop = stop && pop%(id)s->stop_condition();
            else
                stop = stop || pop%(id)s->stop_condition();
    """
}

omp_initialize_template = """
%(prof_init)s
    // Internal variables
    dt = _dt;
    t = static_cast<long int>(0);

    // Populations
%(pop_init)s

    // Projections
%(proj_init)s

    // Custom constants
%(custom_constant)s
"""

cuda_header_template = """#ifndef __ANNARCHY_H__
#define __ANNARCHY_H__

#include <string>
#include <vector>
#include <algorithm>
#include <map>
#include <deque>
#include <queue>
#include <random>
#include <iostream>
#include <sstream>
#include <fstream>
#include <cstdlib>
#include <stdlib.h>
#include <string.h>
#include <cassert>

#include <cuda_runtime_api.h>
#include <curand_kernel.h>

/*
 * Built-in functions (host side)
 */
%(built_in)s

/*
 * Custom constants
 *
 */
%(custom_constant)s

/*
 * Custom functions
 * (available on host-side and interfaced for cython)
 */
%(custom_func)s

/*
 * Structures for the populations
 *
 */
%(pop_struct)s
/*
 * Structures for the projections
 *
 */
%(proj_struct)s

/*
 * Declaration of the populations
 *
 */
%(pop_ptr)s

/*
 * Declaration of the projections
 *
 */
%(proj_ptr)s

/*
 * Recorders
 *
 */
#include "Monitor.hpp"

extern std::vector<Monitor*> recorders;
int addRecorder(Monitor* recorder);
Monitor* getRecorder(int id);
void removeRecorder(Monitor* recorder);

/*
 * Simulation methods
 */
void run(int nbSteps);

int run_until(int steps, std::vector<int> populations, bool or_and);

void step();

/*
 *  Initialization
 */
void initialize(const %(float_prec)s _dt) ;

inline void setDevice(const int device_id) {
#ifdef _DEBUG
    std::cout << "Setting device " << device_id << " as compute device ..." << std::endl;
#endif
    cudaError_t err = cudaSetDevice(device_id);
    if ( err != cudaSuccess )
        std::cerr << "Set device " << device_id << ": " << cudaGetErrorString(err) << std::endl;
}

/*
 *  Life-time management
 */
void create_cpp_instances();
void destroy_cpp_instances();

/*
 * Time export
 */
long int getTime();
void setTime(const long int t_);
%(float_prec)s getDt();
void setDt(const %(float_prec)s dt_);

/*
 * Seed for the RNG (host-side!)
 */
void setSeed(const long int seed, const int num_sources, const bool use_seed_seq);

#endif
"""

cuda_device_kernel = """#include "ANNarchyKernel.cuh"

/********************************************************************/
/*  Device kernel definitions                                       */
/********************************************************************/

/****************************************
 * atomicAdd for non-Pascal             *
 ****************************************/
#if !defined(__CUDA_ARCH__) || __CUDA_ARCH__ >= 600
#else
    __device__ double atomicAdd(double* address, double val)
    {
        unsigned long long int* address_as_ull = (unsigned long long int*)address;
        unsigned long long int old = *address_as_ull, assumed;
        do {
            assumed = old;
            old = atomicCAS(address_as_ull, assumed,
                            __double_as_longlong(val +
                            __longlong_as_double(assumed)));
        } while (assumed != old);
        return __longlong_as_double(old);
    }
#endif

/****************************************
 * init random states                   *
 ****************************************/
/*
 *  Each thread gets an unique sequence number (i) and all use the same seed. As highlightet
 *  in section 3.1.1. of the curand documentation this should be enough to get good random numbers
 *
 *  HD(19.7.2019): we need to be careful, that multiple calls to this method need to generate different state sequences.
 */
__global__ void rng_setup_kernel( int N, long long int sequence_offset, curandState* states, unsigned long long seed )
{
    int i = blockIdx.x * blockDim.x + threadIdx.x;

    while( i < N )
    {
        curand_init( seed, i+sequence_offset, 0, &states[ i ] );
        i += blockDim.x * gridDim.x;
    }
}

/****************************************
 * clear psp-related state variables    *
 ****************************************/
__global__ void clear_num_events(unsigned int* num_events) {
    *num_events = 0;
}

__global__ void clear_sum(int num_elem, %(float_prec)s *sum) {
    int j = threadIdx.x + blockIdx.x * blockDim.x;

    while( j < num_elem ) {
        sum[j] = 0.0;
        j += blockDim.x * gridDim.x;
    }
}

/****************************************
 * common/global available functions    *
 ****************************************/
%(common_kernel)s

/****************************************
 * inline functions                     *
 ****************************************/
%(built_in)s

/****************************************
 * custom constants                     *
 ****************************************/
%(custom_constant)s

/****************************************
 * custom functions                     *
 ****************************************/
%(custom_func)s

/****************************************
 * updating neural variables            *
 ****************************************/
%(pop_kernel)s

/****************************************
 * weighted sum kernels                 *
 ****************************************/
%(psp_kernel)s

/****************************************
 * update synapses kernel               *
 ****************************************/
%(syn_kernel)s

/****************************************
 * global operations kernel             *
 ****************************************/
%(glob_ops_kernel)s

/****************************************
 * postevent kernel                     *
 ****************************************/
%(postevent_kernel)s


/********************************************************************/
/*  Device kernel invocations                                       */
/********************************************************************/

// We need to generate different state sequences per kernel call
static long long int sequence_offset=0;

void init_curand_states( int N, curandState* states, unsigned long long seed ) {
    int numThreads = 64;
    int numBlocks = ceil (float(N) / float(numThreads));

    rng_setup_kernel<<< numBlocks, numThreads >>>( N, sequence_offset, states, seed);
    sequence_offset += N;

#ifdef _DEBUG
    cudaError_t err = cudaGetLastError();
    if ( err != cudaSuccess )
        std::cout << "init_curand_state: " << cudaGetErrorString(err) << std::endl;
#endif
}

void call_clear_sum(RunConfig cfg, int num_elem, %(float_prec)s *sum) {
    clear_sum<<<cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream>>>(num_elem, sum);
}

void call_clear_num_events(RunConfig cfg, unsigned int* num_events) {
    clear_num_events<<<cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream>>>(num_events);
}

/****************************************
 * updating neural variables            *
 ****************************************/
 %(pop_invoke_kernel)s

/****************************************
 * weighted sum kernels                 *
 ****************************************/
%(psp_invoke_kernel)s

/****************************************
 * update synapses kernel               *
 ****************************************/
%(syn_invoke_kernel)s

/****************************************
 * global operations kernel             *
 ****************************************/
%(glob_ops_invoke_kernel)s

/****************************************
 * postevent kernel                     *
 ****************************************/
%(postevent_invoke_kernel)s
"""

cuda_device_invoke_header ="""#pragma once
#include <cuda_runtime_api.h>
#include <curand_kernel.h>
#include <float.h>
#include <stdio.h>
#include <iostream>

// Encapsulates the four parameters required for a kernel invocation.
struct RunConfig{
    dim3 nb;
    dim3 tpb;
    int smem_size;
    cudaStream_t stream;

    RunConfig() = default;

    RunConfig(int nb, int tpb, int smem_size, cudaStream_t stream) {
        this->nb = dim3(nb,1,1);
        this->tpb = dim3(tpb,1,1);
        this->smem_size = smem_size;
        this->stream = stream;
    }

    RunConfig(dim3 nb, dim3 tpb, int smem_size, cudaStream_t stream) {
        this->nb = nb;
        this->tpb = tpb;
        this->smem_size = smem_size;
        this->stream = stream;
    }
};

// Pre-defined kernel definitions
void init_curand_states( int N, curandState* states, unsigned long long seed );

void call_clear_sum(RunConfig cfg, int num_elem, %(float_prec)s *sum);
void call_clear_num_events(RunConfig cfg, unsigned int* num_events);

// Model-related kernel definitions
%(invoke_kernel_def)s

"""

cuda_host_body_template =\
"""// ANNarchy-related header
#include "ANNarchy.hpp"
#include "ANNarchyKernel.cuh"

%(prof_include)s
#include <math.h>

// Directly set the number of blocks/number of threads
%(kernel_config)s

// Required by Bell & Garland Kernel
#define MAX_THREADS (30 * 1024)
#define DIVIDE_INTO(x,y) ((x + y - 1)/y)

//
// Handling GPU and CPU rng
//

std::vector<std::mt19937> rng;
unsigned long long global_seed;

void setSeed(const long int seed, const int num_sources, const bool use_seed_seq){
    rng.clear();

    if (num_sources == 1) {
        rng.push_back(std::mt19937(seed));
    }else {
        if ( use_seed_seq ) {
            std::seed_seq seq{seed};
            std::vector<std::uint32_t> seeds(num_sources);
            seq.generate(seeds.begin(), seeds.end());

            for (auto it = seeds.begin(); it != seeds.end(); it++) {
                rng.push_back(std::mt19937(*it));
            }
        } else {
            std::cerr << "Not implemented. " << std::endl;
        }
    }

    rng.shrink_to_fit();

    // store the seed for later usage
    global_seed = static_cast<unsigned long long>(seed);
}

/*
 * Internal data
 */
%(float_prec)s dt;
long int t;

// Populations
%(pop_ptr)s

// Projections
%(proj_ptr)s

// Stream configuration (available for CC > 3.x devices)
// NOTE: if the CC is lower then 3.x modification of stream
//       parameter (4th arg) is automatically ignored by CUDA
%(stream_setup)s

// Initialize the internal data
void initialize(%(float_prec)s _dt) {
%(initialize)s
}

/*
 *  Life-time management
 */
void create_cpp_instances() {
#ifdef _DEBUG
    std::cout << "Instantiate C++ objects ..." << std::endl;
#endif
}

void destroy_cpp_instances() {
#ifdef _DEBUG
    std::cout << "Destroy C++ objects ..." << std::endl;
#endif
}

/*
 * Recorders
 */
std::vector<Monitor*> recorders;
int addRecorder(Monitor* recorder){
    int found = -1;

    for (unsigned int i=0; i<recorders.size(); i++) {
        if (recorders[i] == nullptr) {
            found = i;
            break;
        }
    }

    if (found != -1) {
        // fill a previously cleared slot
        recorders[found] = recorder;
        return found;
    } else {
        recorders.push_back(recorder);
        return recorders.size() - 1;
    }
}

Monitor* getRecorder(int id) {
    if (id < recorders.size())
        return recorders[id];
    else
        return nullptr;
}

void removeRecorder(Monitor* recorder){
    for (unsigned int i=0; i<recorders.size(); i++){
        if (recorders[i] == recorder) {
            // mark position as free
            recorders[i] = nullptr;
            break;
        }
    }
}

/**
 *  Implementation remark (27.02.2015, HD) to: run(int), step() and single_step()
 *
 *  we have two functions in ANNarchy to run simulation code: run(int) and step(). The latter one to
 *  run several steps at once, the other one just a single step. On CUDA I face the problem, that I
 *  propably need to update variables before step() and definitly changed variables after step().
 *  run(int) calls step() normally, if I add transfer codes in step(), run(N) would end up in N
 *  back transfers from GPUs, whereas we only need the last one.
 *  As solution I renamed step() to single_step(), the interface behaves as OMP side and I only
 *  transfer at begin and at end, as planned.
 */

// Step method. Generated by ANNarchy. (analog to step() in OMP)
void single_step()
{
#ifdef _TRACE_SIMULATION_STEPS
    std::cout << "-- evaluate step " << t << " (" << t * dt << " ms) --" << std::endl;
    std::cout << std::flush;
#endif

%(prof_step_pre)s

%(prof_proj_psp_pre)s
    ////////////////////////////////
    // Clear sums
    ////////////////////////////////
%(clear_sums)s

    ////////////////////////////////
    // Presynaptic events
    ////////////////////////////////
%(compute_sums)s

%(prof_proj_psp_post)s

    ////////////////////////////////
    // Recording targets
    ////////////////////////////////
    for(int i=0; i < recorders.size(); i++){
        if (recorders[i])
            recorders[i]->record_targets();
    }

    ////////////////////////////////
    // Update neural variables
    ////////////////////////////////
%(prof_neur_step_pre)s
%(update_neuron)s

%(prof_neur_step_post)s

%(update_FR)s

    ////////////////////////////////
    // Delay outputs
    ////////////////////////////////
%(delay_code)s

    ////////////////////////////////
    // Global operations (min/max/mean)
    ////////////////////////////////
%(update_globalops)s

    ////////////////////////////////
    // Update synaptic variables
    ////////////////////////////////
%(prof_proj_step_pre)s
%(update_synapse)s

%(prof_proj_step_post)s

    ////////////////////////////////
    // Postsynaptic events
    ////////////////////////////////
%(post_event)s

    ////////////////////////////////
    // Recording neural/synaptic variables
    ////////////////////////////////
%(prof_record_pre)s
    for (unsigned int i=0; i < recorders.size(); i++){
        if (recorders[i])
            recorders[i]->record();
    }
%(prof_record_post)s

    ////////////////////////////////
    // Increase internal time
    ////////////////////////////////
    t++;    // host side, provided as argument to kernels

%(prof_step_post)s

#ifdef _TRACE_SIMULATION_STEPS
    std::cout << "-- simulation step " << t << " completed --" << std::endl;
    std::cout << std::flush;
#endif
}

// Simulate the network for the given number of steps
void run(const int nbSteps) {
#ifdef _DEBUG
    std::cout << "host to device transfers." << std::endl;
#endif

%(host_device_transfer)s

    stream_assign();

#ifdef _DEBUG
    std::cout << "simulate " << nbSteps << " steps." << std::endl;
#endif

%(prof_run_pre)s
    // simulation loop
    for(int i=0; i<nbSteps; i++) {
        single_step();
    }
%(prof_run_post)s

    auto err = cudaGetLastError();
    if (err != cudaSuccess) {
        std::cout << "Error occured during simulation: " << cudaGetErrorString(err) << std::endl;
    }

%(device_host_transfer)s

    cudaDeviceSynchronize();
}

int run_until(const int steps, std::vector<int> populations, bool or_and) {
%(run_until)s
}

void step() {
#ifdef _DEBUG
    std::cout << "host to device transfers." << std::endl;
#endif
%(host_device_transfer)s

    stream_assign();

#ifdef _DEBUG
    std::cout << "simulate a single step." << std::endl;
#endif

%(prof_run_pre)s
    single_step();
%(prof_run_post)s

    auto err = cudaGetLastError();
    if (err != cudaSuccess) {
        std::cerr << "An error occured within simulation loop: " << cudaGetErrorString(err) << std::endl;
    }

%(device_host_transfer)s
    
    cudaDeviceSynchronize();
}

/*
 * Access to time and dt
 *
 */
long int getTime() {return t;}
void setTime(const long int t_) { t=t_; }
%(float_prec)s getDt() { return dt;}
void setDt(const %(float_prec)s dt_) { dt=dt_;}
"""

cuda_stream_setup="""
cudaStream_t streams[%(nbStreams)s];

void stream_setup()
{
    for ( int i = 0; i < %(nbStreams)s; i ++ )
    {
        cudaStreamCreate( &streams[i] );
    }
}

void stream_assign()
{
%(pop_assign)s

%(proj_assign)s
}

void stream_destroy()
{
    for ( int i = 0; i < %(nbStreams)s; i ++ )
    {
        // all work finished
        cudaStreamSynchronize( streams[i] );

        // destroy
        cudaStreamDestroy( streams[i] );
    }
}
"""

cuda_initialize_template = """
    dt = _dt;
    t = (long int)(0);

%(prof_init)s

%(pop_init)s

%(proj_init)s

    // global constants
%(custom_constant)s

    // create streams
    stream_setup();
"""

built_in_functions = """
#define positive(x) (x>0.0? x : 0.0)
#define negative(x) (x<0.0? x : 0.0)
#define clip(x, a, b) (x<a? a : (x>b? b :x))
#define modulo(a, b) long(a) % long(b)
#define ite(a, b, c) (a?b:c)
"""

integer_power_cpu="""
// power function for integer exponent
inline %(float_prec)s power(%(float_prec)s x, unsigned int a){
    %(float_prec)s res=x;
    for (unsigned int i=0; i< a-1; i++){
        res *= x;
    }
    return res;
};
"""

integer_power_cuda="""
// power function for integer exponent
__device__ %(float_prec)s power(%(float_prec)s x, unsigned int a) {
    %(float_prec)s res=x;
    for (unsigned int i = 0; i < a-1; i++){
        res *= x;
    }
    return res;
}
"""



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.core.Population import Population
from ANNarchy.core.Neuron import Neuron
from ANNarchy.intern.ConfigManagement import ConfigManager
from ANNarchy.generator.Compiler import extra_libs 
from ANNarchy.intern import Messages

try:
    from PIL import Image
except:
    Messages._warning('The Python Image Library (pillow) is not installed on your system, unable to create ImagePopulations.')
    
import numpy as np

class ImagePopulation(Population):
    """ 
    Rate-coded Population allowing to represent images (png, jpg...) as the firing rate of a population (each neuron represents one pixel).
    
    This extension requires the Python Image Library (`pip install Pillow`).
    
    The extensions has to be explicitly imported:
 
    ```python
    import ANNarchy as ann
    from ANNarchy.extensions.image import ImagePopulation

    net = ann.Network()
    pop = net.create(ImagePopulation(geometry=(480, 640)))
    pop.set_image('image.jpg')
    ```

    About the geometry:
    
    * If the geometry is 2D, it corresponds to the (height, width) of the image. Only the luminance of the pixels will be represented (grayscale image).
    * If the geometry is 3D, the third dimension can be either 1 (grayscale) or 3 (color).
    
    If the third dimension is 3, each will correspond to the RGB values of the pixels.
    
    **Warning:** due to the indexing system of Numpy, a 640*480 image should be fed into a (480, 640) or (480, 640, 3) population.

    :param geometry: population geometry as tuple. It must correspond to the image size and be fixed through the whole simulation.
    :param name: unique name of the population (optional).  
    """
    
    def __init__(self, geometry:tuple, name:str=None, copied:bool=False, net_id=0):
 
        # Check geometry
        if isinstance(geometry, int) or len(geometry)==1:
            Messages._error('The geometry of an ImagePopulation should be 2D (grayscale) or 3D (color).')
            
        if len(geometry)==3 and (geometry[2]!=3 and geometry[2]!=1):
            Messages._error('The third dimension of an ImagePopulation should be either 1 (grayscale) or 3 (color).') 
                        
        if len(geometry)==3 and geometry[2]==1:
            geometry = (int(geometry[0]), int(geometry[1]))
            
        # Create the population     
        Population.__init__(self, geometry = geometry, name=name, neuron = Neuron(parameters="r = 0.0"), copied=copied, net_id=net_id)
    
    def _copy(self, net_id=None):
        "Returns a copy of the population when creating networks. Internal use only."
        return ImagePopulation(geometry=self.geometry, name=self.name, copied=True, net_id=self.net_id if not net_id else net_id)

    def set_image(self, filename:str) -> None:
        """ 
        Sets an image (.png, .jpg or whatever is supported by PIL) into the firing rate of the population.
        
        If the image has a different size from the population, it will be resized.
        
        :param filename: name of the image file.
        """
        try:
            im = Image.open(filename)
        except : # image does not exist
            Messages._error('The image ' + filename + ' does not exist.')
            
        # Resize the image if needed
        (width, height) = (self.geometry[1], self.geometry[0])
        if im.size != (width, height):
            Messages._warning('The image ' + filename + ' does not have the same size '+str(im.size)+' as the population ' + str((width, height)) + '. It will be resized.')
            im = im.resize((width, height))
        
        # Check if only the luminance should be extracted
        if self.dimension == 2 or self.geometry[2] == 1:
            im=im.convert("L")
        
        # Set the rate of the population
        self.r = np.array(im).reshape(self.size)/255.


class VideoPopulation(ImagePopulation):
    """ 
    Rate-coded Population allowing to feed a webcam input into the firing rate of a population (each neuron represents one pixel).
    
    This extension requires the C++ library OpenCV >= 4.0 (apt-get/yum install opencv). ``pkg-config opencv4 --cflags --libs`` should not return an error. `vtk` might additionally have to be installed.
    
    The extensions has to be explicitly imported:

    ```python
    import ANNarchy as ann
    from ANNarchy.extensions.image import VideoPopulation
    
    net = ann.Network()
    pop = net.create(VideoPopulation(geometry=(480, 640)))
    
    net.compile()
    
    pop.start_camera(0)
    
    while(True):
        pop.grab_image()
        net.simulate(10.0)
    ```

    About the geometry:
    
    * If the geometry is 2D, it corresponds to the (height, width) of the image. Only the luminance of the pixels will be represented (grayscale image).
    * If the geometry is 3D, the third dimension can be either 1 (grayscale) or 3 (color).
    
    If the third dimension is 3, each will correspond to the RGB values of the pixels.
    
    **Warning:** due to the indexing system of Numpy, a 640*480 image should be fed into a (480, 640) or (480, 640, 3) population.

    :param geometry: population geometry as tuple. It must be fixed through the whole simulation. If the camera provides images of a different size, it will be resized.
    :param opencv_version: OpenCV version (default=4).
    :param name: unique name of the population (optional). 
    """
    
    def __init__(self, geometry:tuple, opencv_version:str="4", name:str=None, copied:bool=False, net_id=0):
        
        # Create the population     
        ImagePopulation.__init__(self, geometry = geometry, name=name, copied=copied, net_id=net_id)

        self.opencv_version = opencv_version

    def _copy(self, net_id=None):
        "Returns a copy of the population when creating networks. Internal use only."
        return VideoPopulation(geometry=self.geometry, name=self.name, copied=True, net_id=self.net_id if not net_id else net_id)

    def _generate(self):
        "Code generation"      
        # Add corresponding libs to the Makefile
        extra_libs.append('`pkg-config opencv' + str(self.opencv_version) + ' --cflags --libs`')

        # Include opencv
        self._specific_template['include_additional'] = """#include <opencv2/opencv.hpp>
using namespace cv;
"""
        # Class for the camera device
        self._specific_template['struct_additional'] = """
// VideoPopulation
class CameraDeviceCPP : public cv::VideoCapture {
public:
    CameraDeviceCPP (int id, int width, int height, int depth) : cv::VideoCapture(id){
        width_ = width;
        height_ = height;
        depth_ = depth;      
        img_ = std::vector<%(float_prec)s>(width*height*depth, 0.0);
    }
    std::vector<%(float_prec)s> GrabImage(){
        if(isOpened()){
            // Read a new frame from the video
            Mat frame;
            read(frame); 
            // Resize the image
            Mat resized_frame;
            resize(frame, resized_frame, Size(width_, height_) );
            // If depth=1, only luminance
            if(depth_==1){
                // Convert to luminance
                cvtColor(resized_frame, resized_frame, COLOR_BGR2GRAY);
                for(int i = 0; i < resized_frame.rows; i++){
                    for(int j = 0; j < resized_frame.cols; j++){
                        this->img_[j+width_*i] = float(resized_frame.at<uchar>(i, j))/255.0;
                    }
                }
            }
            else{ //BGR
                for(int i = 0; i < resized_frame.rows; i++){
                    for(int j = 0; j < resized_frame.cols; j++){
                        Vec3b intensity = resized_frame.at<Vec3b>(i, j);
                        this->img_[(j+width_*i)*3 + 0] = %(float_prec)s(intensity.val[2])/255.0;
                        this->img_[(j+width_*i)*3 + 1] = %(float_prec)s(intensity.val[1])/255.0;
                        this->img_[(j+width_*i)*3 + 2] = %(float_prec)s(intensity.val[0])/255.0;
                    }
                }
            }            
        }
        return this->img_;
    };

protected:
    // Width and height of the image, depth_ is 1 (grayscale) or 3 (RGB)
    int width_, height_, depth_;
    // Vector of floats for the returned image
    std::vector<%(float_prec)s> img_;
};
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}

        self._specific_template['declare_additional'] = """
    // Camera
    CameraDeviceCPP* camera_;
    void StartCamera(int id, int width, int height, int depth){
        camera_ = new CameraDeviceCPP(id, width, height, depth);
        if(!camera_->isOpened()){
            std::cout << "Error: could not open the camera." << std::endl;   
        }
    };
    void GrabImage(){
        if(camera_->isOpened()){
            r = camera_->GrabImage();   
        }
    };
    void ReleaseCamera(){
        camera_->release(); 
    };
"""

        self._specific_template['update_variables'] = ""

    def start_camera(self, camera_port=0):
        """
        Starts the webcam with the corresponding device (default = 0).
        
        On linux, the camera port corresponds to the number in `/dev/video0`, `/dev/video1`, etc.
        """

        self.cyInstance.start_camera(camera_port, self.geometry[1], self.geometry[0], 3 if self.dimension==3 else 1)
        
    def grab_image(self):
        """
        Grabs one image from the camera and feeds it into the population.
        
        The camera must be first started with:
        
        ```python
        pop.start_camera(0)
        ```
        """
        self.cyInstance.grab_image()
        
    def release(self):
        """
        Releases the camera:
        
        ```python
        pop.release()
        ```
        """
        self.cyInstance.release_camera()

        



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

import numpy as np

from ANNarchy.intern.SpecificPopulation import SpecificPopulation
from ANNarchy.intern.ConfigManagement import ConfigManager
from ANNarchy.intern import Messages
from ANNarchy.core.Population import Population
from ANNarchy.core.Neuron import Neuron

def _rectify(mu, corr, tau):
    """
    Rectifies mu and sigma to ensure the rates are positive.

    This part of the code is adapted from Brian's source code:

    Copyright ENS, INRIA, CNRS
    Authors: Romain Brette (brette@di.ens.fr) and Dan Goodman (goodman@di.ens.fr)
    Licence: CeCILL
    """

    from scipy.special import erf #pylint: disable=no-name-in-module
    from scipy.optimize import newton

    def _rectified_gaussian(mu, sigma):
        """
        Calculates the mean and standard deviation for a rectified Gaussian distribution.
        mu, sigma: parameters of the original distribution
        Returns mur,sigmar: parameters of the rectified distribution
        """
        a = 1. + erf(mu / (sigma * (2 ** .5)))
        mur = (sigma / (2. * np.pi) ** .5) * np.exp(-0.5 * (mu / sigma) ** 2) + .5 * mu * a
        sigmar = ((mu - mur) * mur + .5 * sigma ** 2 * a) ** .5
        return (mur, sigmar)

    mur = mu
    sigmar = (corr * mu / (2. * tau/1000.)) ** .5
    if sigmar == 0 * sigmar: # for unit consistency
        return (mur, sigmar)
    x0 = mur / sigmar
    ratio = lambda u, v:u / v
    f = lambda x:ratio(*_rectified_gaussian(x, 1.)) - x0
    y = newton(f, x0 * 1.1) # Secant method
    new_sigma = mur / (np.exp(-0.5 * y ** 2) / ((2. * np.pi) ** .5) + .5 * y * (1. + erf(y * (2 ** (-.5)))))
    new_mu = y * new_sigma
    return (new_mu, new_sigma)

class HomogeneousCorrelatedSpikeTrains(SpecificPopulation):
    r"""
    Population of spiking neurons following a homogeneous distribution with correlated spike trains.

    The method describing the generation of homogeneous correlated spike trains is described in:

    > Brette, R. (2009). Generation of correlated spike trains. Neural Computation 21(1). <http://romainbrette.fr/WordPress3/wp-content/uploads/2014/06/Brette2008NC.pdf>

    The implementation is based on the one provided by Brian <http://briansimulator.org>.

    To generate correlated spike trains, the population rate of the group of Poisson-like spiking neurons varies following a stochastic differential equation:

    $$\frac{dx}{dt} = \frac{\mu - x}{\tau} + \sigma \, \frac{\xi}{\sqrt{\tau}}$$

    where $\xi$ is a random sample from the standard normal distribution. In short, $x$ will randomly vary around $\mu$ over time, with an amplitude determined by $\sigma$ and a speed determined by $\tau$.

    This doubly stochastic process is called a Cox process or Ornstein-Uhlenbeck process.

    To avoid that x becomes negative, the values of mu and sigma are computed from a rectified Gaussian distribution, parameterized by the desired population rate **rates**, the desired correlation strength **corr** and the time constant **tau**. See Brette's paper for details.

    In short, you should only define the parameters ``rates``, ``corr`` and ``tau``, and let the class compute mu and sigma for you. Changing ``rates``, ``corr`` or ``tau`` after initialization automatically recomputes mu and sigma.

    Example:

    ```python
    import ANNarchy as ann

    net = ann.Network(dt=0.1)

    pop_corr = net.create(
        ann.HomogeneousCorrelatedSpikeTrains(
            geometry=200, 
            rates=10., 
            corr=0.3, 
            tau=10.)
    )

    net.compile()

    net.simulate(1000.)

    pop_corr.rates=30.

    net.simulate(1000.)
    ```

    Alternatively, a schedule can be provided to change automatically the value of `rates` and `corr` (but not `tau`) at the required times (as in `TimedArray` or TimedPoissonPopulation):

    ```python
    pop_corr = net.create(
        ann.HomogeneousCorrelatedSpikeTrains(
            geometry = 200, 
            rates = [10., 30.], 
            corr = [0.3, 0.5], 
            tau = 10.,
            schedule = [0., 1000.]
        )
    )
    ```

    Even when using a schedule, ``corr`` accepts a single constant value. The first value of ``schedule`` must be 0. ``period`` specifies when the schedule "loops" back to its initial value. 

    :param geometry: population geometry as tuple.
    :param rates: rate in Hz of the population (must be a float or a list of float)
    :param corr: total correlation strength (float in [0, 1], or a list)
    :param tau: correlation time constant in ms.
    :param schedule: list of times where new values of ``rates``and ``corr``will be used to computre mu and sigma.
    :param period: time when the array will be reset and start again, allowing cycling over the schedule. Default: no cycling (-1.)
    :param name: unique name of the population (optional).
    :param refractory: refractory period in ms (careful: may break the correlation)
    """
    def __init__(self, 
            geometry: int | tuple[int], 
            rates:float | list[float], 
            corr:float | list[float], 
            tau:float, 
            schedule:list[float] = None, 
            period:float = -1., 
            name:str = None, 
            refractory:float = None, 
            copied:bool = False,
            net_id:int = 0,
        ):

        # schedule should be a list of onset times. 
        if schedule is None:
            schedule = [0.0]
        else:
            schedule = list(schedule)
        
        # rates should be a list with one value per schedule
        if isinstance(rates, (list, np.ndarray)):
            rates = list(rates)
            if len(rates) != len(schedule): 
                Messages._error("TimedHomogeneousCorrelatedSpikeTrains: the rates argument must be of the same size as schedule when provided as a list.")
        else:
            rates = [float(rates) for _ in range(len(schedule))]

        # corr also
        if isinstance(corr, (list, np.ndarray)):
            if len(corr) != len(schedule): 
                Messages._error("TimedHomogeneousCorrelatedSpikeTrains: the corr argument must be of the same size as schedule when provided as a list.")
            corr = list(corr)
        else:
            corr = [float(corr) for _ in range(len(schedule))]

        # Store the arguments
        self.schedule = schedule
        self.rates = rates
        self.corr = corr
        self.tau = tau
        
        # Store refractory
        self.refractory_init = refractory

        # Correction of mu and sigma
        self.mu_list, self.sigma_list = self._correction(self.rates, self.corr, self.tau)

        # Create the neuron
        corr_neuron = Neuron(
            parameters = """
                tau = %(tau)s : population
                mu = %(mu)s : population
                sigma = %(sigma)s : population
            """ % {'tau': tau, 'mu': self.mu_list[0], 'sigma': self.sigma_list[0]},
            equations = """
                x += dt*(mu - x)/tau + sqrt(dt/tau) * sigma * Normal(0., 1.) : population, init=%(mu)s
                p = Uniform(0.0, 1.0) * 1000.0 / dt
            """ % {'mu': self.mu_list[0]},
            spike = "p < x",
            refractory=refractory,
            name="HomogeneousCorrelated",
            description="Homogeneous correlated spike trains."
        )

        SpecificPopulation.__init__(
            self, 
            geometry=geometry, 
            neuron=corr_neuron, 
            name=name, 
            copied=copied, 
            net_id=net_id
        )

        # Initial values
        self.init['schedule'] = schedule
        self.init['rates'] = rates
        self.init['corr'] = corr
        self.init['tau'] = tau
        self.init['period'] = period
        self.init['mu_list'] = self.mu_list
        self.init['sigma_list'] = self.sigma_list
        self.init['mu'] = self.mu_list[0]
        self.init['sigma'] = self.sigma_list[0]


    def _copy(self, net_id=None):
        "Returns a copy of the population when creating networks."
        return HomogeneousCorrelatedSpikeTrains(
            geometry=self.geometry, 
            rates=self.init['rates'], 
            corr=self.init['corr'], 
            tau=self.init['tau'], 
            schedule=self.init['schedule'], 
            period=self.init['period'], 
            name=self.name, 
            refractory=self.refractory_init, 
            copied=True, 
            net_id=self.net_id if net_id is None else net_id)

    def _correction(self, rates, corr, tau):

        # Correction of mu and sigma
        mu_list = []
        sigma_list = []

        for i in range(len(rates)):
            mu, sigma = _rectify(rates[i], corr[i], tau)
            mu_list.append(mu)
            sigma_list.append(sigma)

        return mu_list, sigma_list

    def _generate_st(self):
        """
        adjust code templates for the specific population for single thread.
        """
        self._specific_template['declare_additional'] = """
    // Custom local parameters of a HomogeneousCorrelatedSpikeTrains
    std::vector< int > _schedule; // List of times where new inputs should be set
    
    std::vector< %(float_prec)s > _mu; // buffer holding the data
    std::vector< %(float_prec)s > _sigma; // buffer holding the data
    
    int _period; // Period of cycling
    long int _t; // Internal time
    int _block; // Internal block when inputs are set not at each step
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}

        self._specific_template['access_additional'] = """
    // Custom local parameters of a HomogeneousCorrelatedSpikeTrains
    void set_schedule(std::vector<int> schedule) { _schedule = schedule; }
    std::vector<int> get_schedule() { return _schedule; }

    void set_mu_list(std::vector< %(float_prec)s > buffer) { _mu = buffer; mu = _mu[0]; }
    std::vector< %(float_prec)s > get_mu_list() { return _mu; }

    void set_sigma_list(std::vector< %(float_prec)s > buffer) { _sigma = buffer; sigma = _sigma[0]; }
    std::vector< %(float_prec)s > get_sigma_list() { return _sigma; }

    void set_period(int period) { _period = period; }
    int get_period() { return _period; }

""" % {'float_prec': ConfigManager().get('precision', self.net_id)}

        self._specific_template['init_additional'] = """
        // Initialize counters
        _t = 0;
        _block = 0;
        _period = -1;
"""

        self._specific_template['reset_additional'] ="""
        _t = 0;
        _block = 0;

        r.clear();
        r = std::vector<%(float_prec)s>(size, 0.0);
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}

        self._specific_template['update_variables'] = """
        if(_active){
            // Check if it is time to set the input
            if(_t == _schedule[_block]){
                // Set the data
                mu = _mu[_block];
                sigma = _sigma[_block];
                // Move to the next block
                _block++;
                // If was the last block, go back to the first block
                if (_block == _schedule.size()){
                    _block = 0;
                }
            }

            // If the timedarray is periodic, check if we arrive at that point
            if(_period > -1 && (_t == _period-1)){
                // Reset the counters
                _block=0;
                _t = -1;
            }

            // Always increment the internal time
            _t++;

            //std::cout << "t: " << _t << " mu: " << mu << " sigma: " << sigma << std::endl;
            
            // Generate spikes
            spiked.clear();

            // x += dt*(mu - x)/tau + sqrt(dt/tau) * sigma * Normal(0., 1.)
            x += dt*(mu - x)/tau + rand_0*sigma*sqrt(dt/tau);

            %(float_prec)s _step = 1000.0/dt;

            #pragma omp simd
            for(int i = 0; i < size; i++){

                // p = Uniform(0.0, 1.0) * 1000.0 / dt
                p[i] = _step*rand_1[i];

            }
        } // active
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}

        
        self._specific_template['size_in_bytes'] = """
        // schedule
        size_in_bytes += _schedule.capacity() * sizeof(int);
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}
        

        # Nanobind
        self._specific_template['wrapper'] = f"""
    // TimedArray
    nanobind::class_<PopStruct{self.id}>(m, "pop{self.id}_wrapper")
        // Constructor
        .def(nanobind::init<int, int>())

        // Common attributes
        .def_rw("size", &PopStruct{self.id}::size)
        .def_rw("max_delay", &PopStruct{self.id}::max_delay)

        // Attributes
		.def_rw("r", &PopStruct{self.id}::r)
		.def_rw("tau", &PopStruct{self.id}::tau)
		.def_rw("mu", &PopStruct{self.id}::mu)
		.def_rw("sigma", &PopStruct{self.id}::sigma)
		.def_rw("x", &PopStruct{self.id}::x)
		.def_rw("p", &PopStruct{self.id}::p)

        // Access methods
        .def("set_schedule", &PopStruct{self.id}::set_schedule)
        .def("get_schedule", &PopStruct{self.id}::get_schedule)

        .def("set_mu_list", &PopStruct{self.id}::set_mu_list)
        .def("get_mu_list", &PopStruct{self.id}::get_mu_list)

        .def("set_sigma_list", &PopStruct{self.id}::set_sigma_list)
        .def("get_sigma_list", &PopStruct{self.id}::get_sigma_list)

        .def("set_period", &PopStruct{self.id}::set_period)
        .def("get_period", &PopStruct{self.id}::get_period)

        // Other methods
		.def("compute_firing_rate", &PopStruct{self.id}::compute_firing_rate)
        .def("activate", &PopStruct{self.id}::set_active)
        .def("reset", &PopStruct{self.id}::reset)
        .def("clear", &PopStruct{self.id}::clear);
        """

    def _generate_omp(self):
        """
        adjust code templates for the specific population for openMP.
        """
        self._specific_template['declare_additional'] = """
    // Custom local parameters of a HomogeneousCorrelatedSpikeTrains
    std::vector< int > _schedule; // List of times where new inputs should be set
    
    std::vector< %(float_prec)s > _mu; // buffer holding the data
    std::vector< %(float_prec)s > _sigma; // buffer holding the data
    
    int _period; // Period of cycling
    long int _t; // Internal time
    int _block; // Internal block when inputs are set not at each step
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}

        self._specific_template['access_additional'] = """
    // Custom local parameters of a HomogeneousCorrelatedSpikeTrains
    void set_schedule(std::vector<int> schedule) { _schedule = schedule; }
    std::vector<int> get_schedule() { return _schedule; }

    void set_mu_list(std::vector< %(float_prec)s > buffer) { _mu = buffer; mu = _mu[0]; }
    std::vector< %(float_prec)s > get_mu_list() { return _mu; }

    void set_sigma_list(std::vector< %(float_prec)s > buffer) { _sigma = buffer; sigma = _sigma[0]; }
    std::vector< %(float_prec)s > get_sigma_list() { return _sigma; }

    void set_period(int period) { _period = period; }
    int get_period() { return _period; }

""" % {'float_prec': ConfigManager().get('precision', self.net_id)}

        self._specific_template['init_additional'] = """
        // Initialize counters
        _t = 0;
        _block = 0;
        _period = -1;
"""

        self._specific_template['reset_additional'] ="""
        _t = 0;
        _block = 0;

        r.clear();
        r = std::vector<%(float_prec)s>(size, 0.0);
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}


        scheduling_block = """
        if(_active){
            // Check if it is time to set the input
            if(_t == _schedule[_block]){
                // Set the data
                mu = _mu[_block];
                sigma = _sigma[_block];
                // Move to the next block
                _block++;
                // If was the last block, go back to the first block
                if (_block == _schedule.size()){
                    _block = 0;
                }
            }

            // If the timedarray is periodic, check if we arrive at that point
            if(_period > -1 && (_t == _period-1)){
                // Reset the counters
                _block=0;
                _t = -1;
            }

            // Always increment the internal time
            _t++;
        }
        """

        update_block = """
        if( _active ) {
            #pragma omp single
            {
                spiked.clear();

                // x += dt*(mu - x)/tau + sqrt(dt/tau) * sigma * Normal(0., 1.)
                x += dt*(mu - x)/tau + rand_0*sigma*sqrt(dt/tau);

                %(float_prec)s _step = 1000.0/dt;

                #pragma omp simd
                for(int i = 0; i < size; i++){

                    // p = Uniform(0.0, 1.0) * 1000.0 / dt
                    p[i] = _step*rand_1[i];

                }
            }
        } // active
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}

        if self._has_schedule:
            self._specific_template['update_variables'] = scheduling_block + update_block
        else:
            self._specific_template['update_variables'] = update_block
        
        self._specific_template['size_in_bytes'] = """
        // schedule
        size_in_bytes += _schedule.capacity() * sizeof(int);
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}

    def _generate_cuda(self):
        """
        Code generation if the CUDA paradigm is set.
        """
        #
        # Code for handling the buffer and schedule parameters
        self._specific_template['declare_additional'] = """
    // Custom local parameter HomogeneousCorrelatedSpikeTrains
    std::vector< int > _schedule;

    std::vector<%(float_prec)s> mu_buffer;      // buffer
    std::vector<%(float_prec)s> sigma_buffer;   // buffer

    int _period; // Period of cycling
    long int _t; // Internal time
    int _block; // Internal block when inputs are set not at each step
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}
        
        self._specific_template['access_additional'] = """
    // Custom local parameter HomogeneousCorrelatedSpikeTrains
    void set_schedule(std::vector<int> schedule) { _schedule = schedule; }
    std::vector<int> get_schedule() { return _schedule; }

    void set_mu_list(std::vector< %(float_prec)s > buffer) { mu_buffer = buffer; }
    void set_sigma_list(std::vector< %(float_prec)s > buffer) { sigma_buffer = buffer; }
    std::vector< %(float_prec)s > get_mu_list() { return mu_buffer; }
    std::vector< %(float_prec)s > get_sigma_list() { return sigma_buffer; }

    void set_period(int period) { _period = period; }
    int get_period() { return _period; }
""" % {'float_prec': ConfigManager().get('precision', self.net_id), 'id': self.id}
        
        self._specific_template['init_additional'] = """
        // counters
        _t = 0;
        _block = 0;
        _period = -1;
"""

        self._specific_template['reset_additional'] = """
        // counters
        _t = 0;
        _block = 0;
"""

        if not self._has_schedule:
            # we can use the normal code generation for GPU kernels
            pass

        else:
            self._specific_template['update_variables'] = """
        if(_active) {
            // Check if it is time to set the input
            if(_t == _schedule[_block]){
                // Set the data
                mu = mu_buffer[_block];
                sigma = sigma_buffer[_block];
                // Move to the next block
                _block++;
                // If was the last block, go back to the first block
                if ( _block == _schedule.size() ) {
                    _block = 0;
                }
            }

            // If the timedarray is periodic, check if we arrive at that point
            if( (_period > -1) && (_t == _period-1) ) {
                // Reset the counters
                _block=0;
                _t = -1;
            }

            // Always increment the internal time
            _t++;
        }
"""

            self._specific_template['update_variable_body'] = """
// Updating global variables of population %(id)s
__global__ void cuPop%(id)s_global_step( const long int t, const double dt, const double tau, double mu, double* x, curandState* rand_0, double sigma )
{
    // x += dt*(mu - x)/tau + sqrt(dt/tau) * sigma * Normal(0., 1.)
    x[0] += dt*(mu - x[0])/tau + curand_normal_double( &rand_0[0] )*sigma*sqrt(dt/tau);
}

// Updating local variables of population %(id)s
__global__ void cuPop%(id)s_local_step( const long int t, const double dt, curandState* rand_1, double* x, unsigned int* num_events, int* spiked, long int* last_spike )
{
    int i = threadIdx.x + blockDim.x * blockIdx.x;

    %(float_prec)s step = 1000.0/dt;

    while ( i < %(size)s )
    {
        // p = Uniform(0.0, 1.0) * 1000.0 / dt
        %(float_prec)s p = curand_uniform_double( &rand_1[i] ) * step;

        if (p < x[0]) {
            int pos = atomicAdd ( num_events, 1);
            spiked[pos] = i;
            last_spike[i] = t;
        }

        i += blockDim.x;
    }

    __syncthreads();
}
""" % {
    'id': self.id,
    'size': self.size,
    'float_prec': ConfigManager().get('precision', self.net_id)
}

            self._specific_template['update_variable_header'] = """__global__ void cuPop%(id)s_global_step( const long int t, const double dt, const double tau, double mu, double* x, curandState* rand_0, double sigma );
__global__ void cuPop%(id)s_local_step( const long int t, const double dt, curandState* rand_1, double* x, unsigned int* num_events, int* spiked, long int* last_spike );
""" % {'id': self.id}

            # Please notice, that the GPU kernels can be launched only with one block. Otherwise, the
            # atomicAdd which is called inside the kernel is not working correct (HD: April 1st, 2021)
            self._specific_template['update_variable_call'] = """
    if (pop%(id)s->_active) {
        // Update the scheduling
        pop%(id)s->update();

        // Reset old events
        clear_num_events<<< 1, 1, 0, pop%(id)s->stream >>>(pop%(id)s->gpu_spike_count);
    #ifdef _DEBUG
        cudaError_t err_clear_num_events_%(id)s = cudaGetLastError();
        if (err_clear_num_events_%(id)s != cudaSuccess)
            std::cout << "pop%(id)s_spike_gather: " << cudaGetErrorString(err_clear_num_events_%(id)s) << std::endl;
    #endif

        // compute the value of x based on mu/sigma
        cuPop%(id)s_global_step<<< 1, 1, 0, pop%(id)s->stream >>>(
            t, dt,
            pop%(id)s->tau,
            pop%(id)s->mu,
            pop%(id)s->gpu_x,
            pop%(id)s->gpu_rand_0,
            pop%(id)s->sigma 
        );
        #ifdef _DEBUG
            cudaError_t err_pop%(id)s_global_step = cudaGetLastError();
            if( err_pop%(id)s_global_step != cudaSuccess) {
                std::cout << "pop%(id)s_step: " << cudaGetErrorString(err_pop%(id)s_global_step) << std::endl;
                exit(0);
            }
        #endif

        // Generate new spike events
        cuPop%(id)s_local_step<<< 1, pop%(id)s->_threads_per_block, 0, pop%(id)s->stream >>>(
            t, dt,
            pop%(id)s->gpu_rand_1,
            pop%(id)s->gpu_x,
            pop%(id)s->gpu_spike_count,
            pop%(id)s->gpu_spiked,
            pop%(id)s->gpu_last_spike
        );
    #ifdef _DEBUG
        cudaError_t err_pop_spike_gather_%(id)s = cudaGetLastError();
        if(err_pop_spike_gather_%(id)s != cudaSuccess) {
            std::cout << "pop%(id)s_spike_gather: " << cudaGetErrorString(err_pop_spike_gather_%(id)s) << std::endl;
            exit(0);
        }
    #endif

        // transfer back the spike counter (needed by record)
        cudaMemcpy( &pop%(id)s->spike_count, pop%(id)s->gpu_spike_count, sizeof(unsigned int), cudaMemcpyDeviceToHost);
    #ifdef _DEBUG
        cudaError_t err_pop%(id)s_async_copy = cudaGetLastError();
        if ( err_pop%(id)s_async_copy != cudaSuccess ) {
            std::cout << "record_spike_count: " << cudaGetErrorString(err_pop%(id)s_async_copy) << std::endl;
            exit(0);
        }
    #endif

        // transfer back the spiked array (needed by record)
        if (pop%(id)s->spike_count > 0) {
            cudaMemcpy( pop%(id)s->spiked.data(), pop%(id)s->gpu_spiked, pop%(id)s->spike_count*sizeof(int), cudaMemcpyDeviceToHost);
        #ifdef _DEBUG
            cudaError_t err_pop%(id)s_async_copy2 = cudaGetLastError();
            if ( err_pop%(id)s_async_copy2 != cudaSuccess ) {
                std::cout << "record_spike: " << cudaGetErrorString(err_pop%(id)s_async_copy2) << std::endl;
                exit(0);
            }
        #endif
        }
    }
""" % {'id': self.id}

        self._specific_template['size_in_bytes'] = "//TODO: "

    def _instantiate(self, module):
        # Create the Cython instance
        self.cyInstance = getattr(module, self.class_name+'_wrapper')(self.size, self.max_delay)

    def __setattr__(self, name, value):

        if not hasattr(self, 'initialized'):
            Population.__setattr__(self, name, value)

        elif name == 'schedule':
            if self.initialized:
                # Cast to steps (integer)
                schedule_list = [int(val / ConfigManager().get('dt', self.net_id)) for val in value]
                self.cyInstance.set_schedule( schedule_list )
            else:
                self.init['schedule'] = value

        elif name == 'mu':
            if self.initialized:
                self.cyInstance.mu = value
            else:
                self.init['mu'] = value

        elif name == 'mu_list':
            if self.initialized:
                self.cyInstance.set_mu_list( value )
            else:
                self.init['mu_list'] = value

        elif name == 'sigma':
            if self.initialized:
                self.cyInstance.sigma = value
            else:
                self.init['sigma'] = value

        elif name == 'sigma_list':
            if self.initialized:
                self.cyInstance.set_sigma_list( value )
            else:
                self.init['sigma_list'] = value

        elif name == "period":
            if self.initialized:
                self.cyInstance.set_period(int(value /ConfigManager().get('dt', self.net_id)))
            else:
                self.init['period'] = value

        elif name == 'rates': 
            # Create a list of rates, even if only one value is provided
            if isinstance(value, (np.ndarray, list)):
                value = list(value)
            else:
                value = [float(value)]
            if not len(value) == len(self.schedule):
                Messages._error("HomogeneousCorrelatedSpikeTrains: rates must have the same length as schedule.")

            if self.initialized:
                # Set the attribute
                Population.__setattr__(self, name, value)
                
                # Correction of mu and sigma everytime rates, corr or tau is changed
                try:
                    mu_list, sigma_list = self._correction(value, self.corr, self.tau)
                    self.cyInstance.set_mu_list( mu_list )
                    self.cyInstance.set_sigma_list( sigma_list )

                except Exception as e:
                    print(e)
            else:
                self.init[name] = value

        elif name == 'corr': 
            # Create a list of corrs, even if only one value is provided
            if isinstance(value, (np.ndarray, list)):
                value = list(value)
            else:
                value = [float(value)]
            if not len(value) == len(self.schedule):
                Messages._error("HomogeneousCorrelatedSpikeTrains: corr must have the same length as schedule.")

            if self.initialized:
                # Set the attribute
                Population.__setattr__(self, name, value)
                
                # Correction of mu and sigma everytime rates, corr or tau is changed
                try:
                    mu_list, sigma_list = self._correction(self.rates, value, self.tau)
                    self.cyInstance.set_mu_list( mu_list )
                    self.cyInstance.set_sigma_list( sigma_list )
                except Exception as e:
                    print(e)
            else:
                self.init[name] = value

        elif name == 'tau': 
            if self.initialized:
                # Set the attribute
                Population.__setattr__(self, name, value)
                
                # Correction of mu and sigma everytime rates, corr or tau is changed
                try:
                    mu_list, sigma_list = self._correction(self.rates, self.corr, value)
                    self.cyInstance.set_mu_list( mu_list )
                    self.cyInstance.set_sigma_list( sigma_list )
                except Exception as e:
                    print(e)
            else:
                self.init[name] = value
        else:
            Population.__setattr__(self, name, value)

    def __getattr__(self, name):

        if name == 'schedule':
            if self.initialized:
                return [ConfigManager().get('dt', self.net_id) * val for val in self.cyInstance.get_schedule() ]
            else:
                return self.init['schedule']
              
        elif name == 'mu':
            if self.initialized:
                return self.cyInstance.mu
            else:
                return self.init['mu']
            
        elif name == 'mu_list':
            if self.initialized:
                return list(self.cyInstance.get_mu_list())
            else:
                return self.init['mu_list']
            
        elif name == 'sigma':
            if self.initialized:
                return self.cyInstance.sigma
            else:
                return self.init['sigma']
            
        elif name == 'sigma_list':
            if self.initialized:
                return list(self.cyInstance.get_sigma_list())
            else:
                return self.init['sigma_list']
                      
        elif name == 'period':
            if self.initialized:
                return self.cyInstance.get_period() * ConfigManager().get('dt', self.net_id)
            else:
                return self.init['period']
            
        else:
            return Population.__getattribute__(self, name)



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

cuda_population = {
    'template': """
class PopRecorder%(id)s : public Monitor
{
protected:
    int _id;

public:
    PopRecorder%(id)s(std::vector<int> ranks, int period, int period_offset, long int offset)
        : Monitor(ranks, period, period_offset, offset)
        {
    #ifdef _DEBUG
        std::cout << "PopRecorder%(id)s (" << this << ") instantiated." << std::endl;
    #endif
%(init_code)s

        // add monitor to global list
        this->_id = addRecorder(static_cast<Monitor*>(this));
    #ifdef _DEBUG
        std::cout << "PopRecorder%(id)s (" << this << ") received list position (ID) = " << this->_id << std::endl;
    #endif
    }

    ~PopRecorder%(id)s() {
    #ifdef _DEBUG
        std::cout << "PopRecorder%(id)s::~PopRecorder%(id)s - this = " << this << std::endl;
    #endif
    }

    void record() {
%(recording_code)s
    }

    void record_targets() {
%(recording_target_code)s
    }

    long int size_in_bytes() {
        size_t size_in_bytes = 0;
%(size_in_bytes)s
        return static_cast<long int>(size_in_bytes);
    }

%(clear_individual_container_code)s

    void clear() {
    #ifdef _DEBUG
        std::cout << "PopRecorder%(id)s::clear() - this = " << this << std::endl;
    #endif
%(clear_all_container_code)s

        removeRecorder(this);
    }

%(struct_code)s
};
""",
    'local': {
    'struct': """
    // Local variable %(name)s
    std::vector< std::vector< %(type)s > > %(name)s ;
    bool record_%(name)s ; """,
    'init': """
        this->%(name)s = std::vector< std::vector< %(type)s > >();
        this->record_%(name)s = false; """,
    'recording': """
        if(this->record_%(name)s && ( (t - this->offset_) %% this->period_ == this->period_offset_ )){
            cudaMemcpy(pop%(id)s->%(name)s.data(), pop%(id)s->gpu_%(name)s, pop%(id)s->size * sizeof(%(type)s), cudaMemcpyDeviceToHost);
        #ifdef _DEBUG
            auto err = cudaGetLastError();
            if ( err != cudaSuccess ) {
                std::cout << "record %(name)s on pop%(id)s failed: " << cudaGetErrorString(err) << std::endl;
            } else {
                std::cout << "record %(name)s - [min, max]: " << *std::min_element(pop%(id)s->%(name)s.begin(), pop%(id)s->%(name)s.end() ) << ", " << *std::max_element(pop%(id)s->%(name)s.begin(), pop%(id)s->%(name)s.end() ) << std::endl;
            }
        #endif
            if(!this->partial)
                this->%(name)s.push_back(pop%(id)s->%(name)s);
            else{
                std::vector<%(type)s> tmp = std::vector<%(type)s>();
                for (unsigned int i=0; i<this->ranks.size(); i++){
                    tmp.push_back(pop%(id)s->%(name)s[this->ranks[i]]);
                }
                this->%(name)s.push_back(tmp);
            }
        }""",
    'clear': """
    void clear_%(name)s() {
        for(auto it = this->%(name)s.begin(); it != this->%(name)s.end(); it++) {
            it->clear();
            it->shrink_to_fit();
        }
        this->%(name)s.clear();
    }
""",
    'size_in_bytes': ""
    },
    'semiglobal': { # Does not exist for populations
        'struct': "",
        'init': "",
        'recording': "",
        'clear': "",
        'size_in_bytes': ""
    },
    'global': {
    'struct': """
    // Global variable %(name)s
    std::vector< %(type)s > %(name)s ;
    bool record_%(name)s ; """,
    'init': """
        this->%(name)s = std::vector< %(type)s >();
        this->record_%(name)s = false; """,
    'recording': """
        if(this->record_%(name)s && ( (t - this->offset_) %% this->period_ == this->period_offset_ )){
            this->%(name)s.push_back(pop%(id)s->%(name)s);
        } """,
    'clear': """
    void clear_%(name)s() {
        this->%(name)s.clear();
    }
    """,
    'size_in_bytes': ""
    }
}

cuda_projection = {
    'struct': """
class ProjRecorder%(id)s : public Monitor
{
protected:
    int _id;
    std::vector<int> indices;

public:
    ProjRecorder%(id)s(std::vector<int> ranks, int period, int period_offset, long int offset)
        : Monitor(ranks, period, period_offset, offset)
    {
    #ifdef _DEBUG
        std::cout << "ProjRecorder%(id)s (" << this << ") instantiated." << std::endl;
    #endif
        std::map< int, int > post_indices = std::map< int, int > ();
        auto post_rank = proj%(id)s->get_post_rank();

        for(int i=0; i<post_rank.size(); i++){
            post_indices[post_rank[i]] = i;
        }
        for(int i=0; i<this->ranks.size(); i++){
            this->indices.push_back(post_indices[this->ranks[i]]);
        }
        post_indices.clear();

        // initialize container    
%(init_code)s

        // add monitor to global list
        this->_id = addRecorder(static_cast<Monitor*>(this));
    #ifdef _DEBUG
        std::cout << "ProjRecorder%(id)s (" << this << ") received list position (ID) = " << this->_id << std::endl;
    #endif
    };

    void record() {
%(recording_code)s
    };

    void record_targets() { /* nothing to do here */ }
    long int size_in_bytes() {
        std::cout << "ProjMonitor::size_in_bytes(): not implemented for cuda paradigm." << std::endl;
        return 0;
    }

    void clear() {
    #ifdef _DEBUG
        std::cout << "ProjMonitor%(id)s::clear()." << std::endl;
    #endif
%(clear_container_code)s
    }

%(clear_individual_code)s

%(struct_code)s
};
""",
    'local': {
        'struct': """
    // Local variable %(name)s
    std::vector< std::vector< std::vector< %(type)s > > > %(name)s ;
    bool record_%(name)s ;
""",
    'init': """
        // Local variable %(name)s
        this->%(name)s = std::vector< std::vector< std::vector< %(type)s > > >();
        this->record_%(name)s = false;
""",
    'recording': """
        // Local variable %(name)s
        if(this->record_%(name)s && ( (t - this->offset_) %% this->period_ == this->period_offset_ )){
            auto host_local = proj%(id)s->get_device_matrix_variable_as_lil<%(float_prec)s>(proj%(id)s->gpu_%(name)s);

            std::vector< std::vector< double > > tmp;
            for (auto idx = 0; idx < this->ranks.size(); idx++) {
                tmp.push_back(host_local[this->indices[idx]]);
            }

            this->%(name)s.push_back(tmp);
            tmp.clear();
        }
""",
    'clear': """
    void clear_%(name)s() {
        // local variable %(name)s
        for (auto it = %(name)s.begin(); it != %(name)s.end(); it++) {
            it->clear();
            it->shrink_to_fit();
        }
        %(name)s.clear();
        %(name)s.shrink_to_fit();
    }
""",
        'size_in_bytes': """
// local variable %(name)s
size_in_bytes += sizeof(std::vector<%(type)s>) * %(name)s.capacity();
for (auto it = %(name)s.begin(); it != %(name)s.end(); it++) {
    size_in_bytes += sizeof(%(type)s) * it->capacity();
}
"""
    },
    'semiglobal': {
        'struct': """
    // Semiglobal variable %(name)s
    std::vector< std::vector< %(type)s > > %(name)s ;
    bool record_%(name)s ;
""",
    'init': """
        // Semiglobal variable %(name)s
        this->%(name)s = std::vector< std::vector< %(type)s > > ( this->ranks.size(), std::vector<%(type)s>() );
        this->record_%(name)s = false;
""",
    'recording': """
        // Semiglobal variable %(name)s
        if(this->record_%(name)s && ( (t - this->offset_) %% this->period_ == this->period_offset_ ) ) {
            auto data = std::vector<%(type)s>(proj%(id)s->nb_dendrites(), 0.0);
            cudaMemcpy( data.data(), proj%(id)s->gpu_%(name)s, proj%(id)s->nb_dendrites() * sizeof(%(type)s), cudaMemcpyDeviceToHost);

        #ifdef _DEBUG
            auto err = cudaGetLastError();
            if ( err != cudaSuccess )
                std::cout << "record %(name)s on proj%(id)s failed: " << cudaGetErrorString(err) << std::endl;
        #endif

            for ( int i = 0; i < this->ranks.size(); i++) {
                this->%(name)s[i].push_back(data[this->ranks[i]]);
            }
        }
""",
        'clear': """
    void clear_%(name)s() {
        // semiglobal variable %(name)s
        for (auto it = %(name)s.begin(); it != %(name)s.end(); it++) {
            it->clear();
            it->shrink_to_fit();
        }
        %(name)s.clear();
        %(name)s.shrink_to_fit();
    }
""",
        'size_in_bytes': """
// semiglobal variable %(name)s
size_in_bytes += sizeof(std::vector<%(type)s>) * %(name)s.capacity();
for (auto it = %(name)s.begin(); it != %(name)s.end(); it++) {
    size_in_bytes += sizeof(%(type)s) * it->capacity();
}
"""
    },
    'global': {
        'struct': """
    // Global variable %(name)s
    std::vector< %(type)s > %(name)s ;
    bool record_%(name)s ;
""",
    'init': """
        // Global variable %(name)s
        this->%(name)s = std::vector< %(type)s >();
        this->record_%(name)s = false;
""",
    'recording': """
        // Global variable %(name)s
        if(this->record_%(name)s && ( (t - this->offset_) %% this->period_ == this->period_offset_ )){
            %(type)s tmp = %(type)s(0);
            cudaMemcpy( &tmp, proj%(id)s->gpu_%(name)s, sizeof(%(type)s), cudaMemcpyDeviceToHost);

            this->%(name)s.push_back(tmp);
        #ifdef _DEBUG
            auto err = cudaGetLastError();
            if ( err != cudaSuccess )
                std::cout << "record %(name)s on proj%(id)s failed: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
""",
        'clear': """
    void clear_%(name)s() {
        // global variable %(name)s
        %(name)s.clear();
        %(name)s.shrink_to_fit();
    }
""",
        'size_in_bytes': """
// global variable %(name)s
size_in_bytes += sizeof(%(type)s) * %(name)s.capacity();
"""
    }
}
